from __future__ import annotations

import builtins
import ctypes
from typing import Any
from typing import cast

import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.libc
from pwndbg.lib.ctypes import Structure


def request2size(req: int) -> int:
    if req + SIZE_SZ + MALLOC_ALIGN_MASK < MINSIZE:
        return MINSIZE
    return (req + SIZE_SZ + MALLOC_ALIGN_MASK) & ~MALLOC_ALIGN_MASK


def fastbin_index(size: int) -> int:
    if pwndbg.aglib.arch.ptrsize == 8:
        return (size >> 4) - 2
    return (size >> 3) - 2


# Operating under the assumption that the pwndbg/libc/ code can figure out
# that we are using glibc with at least as good accuracy as the ptmalloc code.
assert pwndbg.libc.which() == pwndbg.libc.LibcType.GLIBC
GLIBC_VERSION = pwndbg.libc.version()
# TODO: Move these heap constants and macros to elsewhere, because pwndbg/aglib/heap/ptmalloc.py also uses them, we are duplicating them here.
SIZE_SZ = pwndbg.aglib.arch.ptrsize
MINSIZE = pwndbg.aglib.arch.ptrsize * 4
if pwndbg.aglib.arch.name == "i386" and GLIBC_VERSION >= (2, 26):
    # i386 will override it to 16 when GLIBC version >= 2.26
    # See https://elixir.bootlin.com/glibc/glibc-2.26/source/sysdeps/i386/malloc-alignment.h#L22
    MALLOC_ALIGN = 16
else:
    # See https://elixir.bootlin.com/glibc/glibc-2.37/source/sysdeps/generic/malloc-alignment.h#L27
    long_double_alignment = pwndbg.aglib.typeinfo.lookup_types("long double").alignof
    MALLOC_ALIGN = long_double_alignment if 2 * SIZE_SZ < long_double_alignment else 2 * SIZE_SZ

MALLOC_ALIGN_MASK = MALLOC_ALIGN - 1
MAX_FAST_SIZE = 80 * SIZE_SZ // 4
NBINS = 128
BINMAPSIZE = 4
TCACHE_SMALL_BINS = 64
TCACHE_LARGE_BINS = 12
TCACHE_MAX_BINS = TCACHE_SMALL_BINS + TCACHE_LARGE_BINS
NFASTBINS = fastbin_index(request2size(MAX_FAST_SIZE)) + 1

if pwndbg.aglib.arch.ptrsize == 4:
    PTR = ctypes.c_uint32
    SIZE_T = ctypes.c_uint32
else:
    PTR = ctypes.c_uint64  # type: ignore[misc]
    SIZE_T = ctypes.c_uint64  # type: ignore[misc]

DEFAULT_TOP_PAD = 131072
DEFAULT_MMAP_MAX = 65536
DEFAULT_MMAP_THRESHOLD = 128 * 1024
DEFAULT_TRIM_THRESHOLD = 128 * 1024
DEFAULT_PAGE_SIZE = 4096
TCACHE_FILL_COUNT = 7
MAX_TCACHE_SMALL_SIZE = (TCACHE_SMALL_BINS - 1) * MALLOC_ALIGN + MINSIZE - SIZE_SZ


class c_pvoid(PTR):
    """
    Represents a pointer.
    """


class c_size_t(SIZE_T):
    """
    Represents a size_t.
    """


C2GDB_MAPPING = {
    ctypes.c_char: pwndbg.aglib.typeinfo.char,
    ctypes.c_int8: pwndbg.aglib.typeinfo.int8,
    ctypes.c_int16: pwndbg.aglib.typeinfo.int16,
    ctypes.c_int32: pwndbg.aglib.typeinfo.int32,
    ctypes.c_int64: pwndbg.aglib.typeinfo.int64,
    ctypes.c_uint8: pwndbg.aglib.typeinfo.uint8,
    ctypes.c_uint16: pwndbg.aglib.typeinfo.uint16,
    ctypes.c_uint32: pwndbg.aglib.typeinfo.uint32,
    ctypes.c_uint64: pwndbg.aglib.typeinfo.uint64,
    c_pvoid: pwndbg.aglib.typeinfo.pvoid,
    c_size_t: pwndbg.aglib.typeinfo.size_t,
}

# Use correct endian for the dictionary keys
if pwndbg.aglib.arch.endian == "little":
    C2GDB_MAPPING: dict[type[ctypes.c_char], pwndbg.dbg_mod.Type] = {  # type: ignore[no-redef]
        k.__ctype_le__: v for k, v in C2GDB_MAPPING.items()
    }
else:
    C2GDB_MAPPING: dict[type[ctypes.c_char], pwndbg.dbg_mod.Type] = {  # type: ignore[no-redef]
        k.__ctype_be__: v for k, v in C2GDB_MAPPING.items()
    }


class FakeGDBField:
    """
    Fake gdb.Field for compatibility
    """

    def __init__(
        self,
        bitpos: int,
        name: str | None,
        type,
        parent_type,
        enumval: int | None = None,
        artificial: bool = False,
        is_base_class: bool = False,
        bitsize: int = 0,
    ) -> None:
        # Note: pwndbg only uses `name` currently
        self.bitpos = bitpos
        self.name = name
        self.type = type
        self.parent_type = parent_type
        if enumval:
            self.enumval = enumval
        self.artificial = artificial
        self.is_base_class = is_base_class
        self.bitsize = bitsize


class CStruct2GDB:
    code = pwndbg.dbg_mod.TypeCode.STRUCT
    _c_struct: builtins.type[ctypes.Structure]

    def __init__(self, address: int) -> None:
        self.address = address

    def __int__(self) -> int:
        """
        Returns the address of the C struct.
        """
        return self.address

    def __getitem__(self, key: str) -> pwndbg.dbg_mod.Value:
        """
        Returns the value of the specified field as a `pwndbg.dbg_mod.Value`.
        """
        return self.read_field(key)

    def __getattr__(self, key: str) -> pwndbg.dbg_mod.Value:
        """
        Returns the value of the specified field as a `pwndbg.dbg_mod.Value`.
        """
        return self.read_field(key)

    def __eq__(self, other: Any) -> bool:
        return self.address == int(other)

    def value_to_human_readable(self) -> str:
        """
        Returns a string representation of the C struct like `pwndbg.dbg_mod.Value` does.
        """
        output = "{\n"
        for f in self._c_struct._fields_:
            output += f"  {f[0]} = {self.read_field(f[0]).inner},\n"
        output += "}"
        return output

    def read_field(self, field: str) -> pwndbg.dbg_mod.Value:
        """
        Returns the value of the specified field as a `pwndbg.dbg_mod.Value`.
        """
        field_address = self.get_field_address(field)
        field_type = next(f for f in self._c_struct._fields_ if f[0] == field)[1]
        if hasattr(field_type, "_length_"):  # f is a ctypes Array
            array_type = cast(ctypes.Array[Any], field_type)
            t = C2GDB_MAPPING[array_type._type_]
            return pwndbg.aglib.memory.get_typed_pointer_value(
                t.array(array_type._length_), field_address
            )
        return pwndbg.aglib.memory.get_typed_pointer_value(C2GDB_MAPPING[field_type], field_address)

    @property
    def type(self):
        """
        Returns type(self) to make it compatible with the `pwndbg.dbg_mod.Value` interface.
        """
        return type(self)

    @classmethod
    def unqualified(cls):
        """
        Returns cls to make it compatible with the `gdb.types.has_field()` interface.
        """
        return cls

    @classmethod
    def fields(cls) -> list[FakeGDBField]:
        """
        Return fields of the struct to make it compatible with the `pwndbg.dbg_mod.Type` interface.
        """
        fake_gdb_fields: list[FakeGDBField] = []
        for f in cls._c_struct._fields_:
            field_name = f[0]
            field_type = f[1]
            bitpos = getattr(cls._c_struct, field_name).offset * 8
            if hasattr(field_type, "_length_"):  # f is a ctypes Array
                arr_type = cast(ctypes.Array[Any], field_type)
                t = C2GDB_MAPPING[arr_type._type_]
                _type = t.array(arr_type._length_)
            else:
                _type = C2GDB_MAPPING[field_type]
            fake_gdb_fields.append(FakeGDBField(bitpos, field_name, _type, cls))
        return fake_gdb_fields

    @classmethod
    def keys(cls) -> list[str]:
        """
        Return a list of the names of the fields in the struct to make it compatible with the `pwndbg.dbg_mod.Type` interface.
        """
        return [f[0] for f in cls._c_struct._fields_]

    def get_field_address(self, field: str) -> int:
        """
        Returns the address of the specified field.
        """
        return self.address + getattr(self._c_struct, field).offset

    @classmethod
    def get_field_offset(cls, field: str) -> int:
        """
        Returns the offset of the specified field.
        """
        return getattr(cls._c_struct, field).offset

    def items(self) -> tuple[tuple[Any, Any], ...]:
        """
        Returns a tuple of (field name, field value) pairs.
        """
        return tuple((field[0], getattr(self, field[0])) for field in self._c_struct._fields_)

    @classmethod
    def has_field(self, field: str) -> bool:
        """
        Checks whether a field exists to make it compatible with the `pwndbg.dbg_mod.Type` interface.
        """
        return field in self.keys()


class c_malloc_state_2_26(Structure):
    """
    This class represents malloc_state struct for GLIBC < 2.27 as a ctypes struct.

    https://github.com/bminor/glibc/blob/1c9a5c270d8b66f30dcfaf1cb2d6cf39d3e18369/malloc/malloc.c#L1678-L1716

    struct malloc_state
    {
        /* Serialize access.  */
        __libc_lock_define (, mutex);

        /* Flags (formerly in max_fast).  */
        int flags;

        /* Fastbins */
        mfastbinptr fastbinsY[NFASTBINS];

        /* Base of the topmost chunk -- not otherwise kept in a bin */
        mchunkptr top;

        /* The remainder from the most recent split of a small request */
        mchunkptr last_remainder;

        /* Normal bins packed as described above */
        mchunkptr bins[NBINS * 2 - 2];

        /* Bitmap of bins */
        unsigned int binmap[BINMAPSIZE];

        /* Linked list */
        struct malloc_state *next;

        /* Linked list for free arenas.  Access to this field is serialized
            by free_list_lock in arena.c.  */
        struct malloc_state *next_free;

        /* Number of threads attached to this arena.  0 if the arena is on
            the free list.  Access to this field is serialized by
            free_list_lock in arena.c.  */
        INTERNAL_SIZE_T attached_threads;

        /* Memory allocated from the system in this arena.  */
        INTERNAL_SIZE_T system_mem;
        INTERNAL_SIZE_T max_system_mem;
    };
    """

    _fields_ = [
        ("mutex", ctypes.c_int32),
        ("flags", ctypes.c_int32),
        ("fastbinsY", c_pvoid * NFASTBINS),
        ("top", c_pvoid),
        ("last_remainder", c_pvoid),
        ("bins", c_pvoid * (NBINS * 2 - 2)),
        ("binmap", ctypes.c_int32 * BINMAPSIZE),
        ("next", c_pvoid),
        ("next_free", c_pvoid),
        ("attached_threads", c_size_t),
        ("system_mem", c_size_t),
        ("max_system_mem", c_size_t),
    ]


class c_malloc_state_2_12(Structure):
    """
    This class represents malloc_state struct for GLIBC < 2.22 as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.12/malloc/malloc.c#L2362-L2400

    struct malloc_state {
    /* Serialize access.  */
    mutex_t mutex;

    /* Flags (formerly in max_fast).  */
    int flags;

    #if THREAD_STATS
    /* Statistics for locking.  Only used if THREAD_STATS is defined.  */
    long stat_lock_direct, stat_lock_loop, stat_lock_wait;
    #endif

    /* Fastbins */
    mfastbinptr      fastbinsY[NFASTBINS];

    /* Base of the topmost chunk -- not otherwise kept in a bin */
    mchunkptr        top;

    /* The remainder from the most recent split of a small request */
    mchunkptr        last_remainder;

    /* Normal bins packed as described above */
    mchunkptr        bins[NBINS * 2 - 2];

    /* Bitmap of bins */
    unsigned int     binmap[BINMAPSIZE];

    /* Linked list */
    struct malloc_state *next;

    #ifdef PER_THREAD
    /* Linked list for free arenas.  */
    struct malloc_state *next_free;
    #endif

    /* Memory allocated from the system in this arena.  */
    INTERNAL_SIZE_T system_mem;
    INTERNAL_SIZE_T max_system_mem;
    };"""

    _fields_ = [
        ("mutex", ctypes.c_int32),
        ("flags", ctypes.c_int32),
        ("fastbinsY", c_pvoid * NFASTBINS),
        ("top", c_pvoid),
        ("last_remainder", c_pvoid),
        ("bins", c_pvoid * (NBINS * 2 - 2)),
        ("binmap", ctypes.c_int32 * BINMAPSIZE),
        ("next", c_pvoid),
        ("next_free", c_pvoid),
        ("system_mem", c_size_t),
        ("max_system_mem", c_size_t),
    ]


class c_malloc_state_2_27(Structure):
    """
    This class represents malloc_state struct for GLIBC >= 2.27 as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.34/malloc/malloc.c#L1831


    struct malloc_state
    {
        /* Serialize access.  */
        __libc_lock_define (, mutex);

        /* Flags (formerly in max_fast).  */
        int flags;

        /* Set if the fastbin chunks contain recently inserted free blocks.  */
        /* Note this is a bool but not all targets support atomics on booleans.  */
        int have_fastchunks;

        /* Fastbins */
        mfastbinptr fastbinsY[NFASTBINS];

        /* Base of the topmost chunk -- not otherwise kept in a bin */
        mchunkptr top;

        /* The remainder from the most recent split of a small request */
        mchunkptr last_remainder;

        /* Normal bins packed as described above */
        mchunkptr bins[NBINS * 2 - 2];

        /* Bitmap of bins */
        unsigned int binmap[BINMAPSIZE];

        /* Linked list */
        struct malloc_state *next;

        /* Linked list for free arenas.  Access to this field is serialized
            by free_list_lock in arena.c.  */
        struct malloc_state *next_free;

        /* Number of threads attached to this arena.  0 if the arena is on
            the free list.  Access to this field is serialized by
            free_list_lock in arena.c.  */
        INTERNAL_SIZE_T attached_threads;

        /* Memory allocated from the system in this arena.  */
        INTERNAL_SIZE_T system_mem;
        INTERNAL_SIZE_T max_system_mem;
    };
    """

    _fields_ = [
        ("mutex", ctypes.c_int32),
        ("flags", ctypes.c_int32),
        ("have_fastchunks", ctypes.c_int32),
        ("fastbinsY", c_pvoid * NFASTBINS),
        ("top", c_pvoid),
        ("last_remainder", c_pvoid),
        ("bins", c_pvoid * (NBINS * 2 - 2)),
        ("binmap", ctypes.c_int32 * BINMAPSIZE),
        ("next", c_pvoid),
        ("next_free", c_pvoid),
        ("attached_threads", c_size_t),
        ("system_mem", c_size_t),
        ("max_system_mem", c_size_t),
    ]


class MallocState(CStruct2GDB):
    """
    This class represents malloc_state struct with interface compatible with `pwndbg.dbg_mod.Value`.
    """

    if GLIBC_VERSION >= (2, 27):
        _c_struct = c_malloc_state_2_27
    elif GLIBC_VERSION >= (2, 23):
        _c_struct = c_malloc_state_2_26
    else:
        _c_struct = c_malloc_state_2_12
    sizeof = ctypes.sizeof(_c_struct)


class c_heap_info(Structure):
    """
    This class represents heap_info struct as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.34/malloc/arena.c#L53

    typedef struct _heap_info
    {
        mstate ar_ptr; /* Arena for this heap. */
        struct _heap_info *prev; /* Previous heap. */
        size_t size;   /* Current size in bytes. */
        size_t mprotect_size; /* Size in bytes that has been mprotected
                                PROT_READ|PROT_WRITE.  */
        /* Make sure the following data is properly aligned, particularly
            that sizeof (heap_info) + 2 * SIZE_SZ is a multiple of
            MALLOC_ALIGNMENT. */
        char pad[-6 * SIZE_SZ & MALLOC_ALIGN_MASK];
    } heap_info;
    """

    _fields_ = [
        ("ar_ptr", c_pvoid),
        ("prev", c_pvoid),
        ("size", c_size_t),
        ("mprotect_size", c_size_t),
        ("pad", ctypes.c_uint8 * (-6 * SIZE_SZ & MALLOC_ALIGN_MASK)),
    ]


class HeapInfo(CStruct2GDB):
    """
    This class represents heap_info struct with interface compatible with `pwndbg.dbg_mod.Value`.
    """

    _c_struct = c_heap_info
    sizeof = ctypes.sizeof(_c_struct)


class c_malloc_chunk(Structure):
    """
    This class represents malloc_chunk struct as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.34/malloc/malloc.c#L1154

    struct malloc_chunk {

        INTERNAL_SIZE_T      mchunk_prev_size;  /* Size of previous chunk (if free).  */
        INTERNAL_SIZE_T      mchunk_size;       /* Size in bytes, including overhead. */

        struct malloc_chunk* fd;         /* double links -- used only if free. */
        struct malloc_chunk* bk;

        /* Only used for large blocks: pointer to next larger size.  */
        struct malloc_chunk* fd_nextsize; /* double links -- used only if free. */
        struct malloc_chunk* bk_nextsize;
    };
    """

    _fields_ = [
        ("prev_size", c_size_t),
        ("size", c_size_t),
        ("fd", c_pvoid),
        ("bk", c_pvoid),
        ("fd_nextsize", c_pvoid),
        ("bk_nextsize", c_pvoid),
    ]


class MallocChunk(CStruct2GDB):
    """
    This class represents malloc_chunk struct with interface compatible with `pwndbg.dbg_mod.Value`.
    """

    _c_struct = c_malloc_chunk
    sizeof = ctypes.sizeof(_c_struct)


class c_tcache_perthread_struct_2_29(Structure):
    """
    This class represents tcache_perthread_struct for GLIBC < 2.30 as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.29/malloc/malloc.c#L2916

    typedef struct tcache_perthread_struct
    {
        char counts[TCACHE_MAX_BINS];
        tcache_entry *entries[TCACHE_MAX_BINS];
    } tcache_perthread_struct;
    """

    _fields_ = [
        ("counts", ctypes.c_char * TCACHE_SMALL_BINS),
        ("entries", c_pvoid * TCACHE_SMALL_BINS),
    ]


class c_tcache_perthread_struct_2_30(Structure):
    """
    This class represents the tcache_perthread_struct for 2.30 <= GLIBC < 2.42 as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.34/malloc/malloc.c#L3025

    typedef struct tcache_perthread_struct
    {
        uint16_t counts[TCACHE_MAX_BINS];
        tcache_entry *entries[TCACHE_MAX_BINS];
    } tcache_perthread_struct;
    """

    _fields_ = [
        ("counts", ctypes.c_uint16 * TCACHE_SMALL_BINS),
        ("entries", c_pvoid * TCACHE_SMALL_BINS),
    ]


class c_tcache_perthread_struct_2_42(Structure):
    """
    This class represents the tcache_perthread_struct for 2.42 <= GLIBC as a ctypes struct.

    https://elixir.bootlin.com/glibc/glibc-2.42/source/malloc/malloc.c#L3127

    typedef struct tcache_perthread_struct
    {
        uint16_t num_slots[TCACHE_MAX_BINS];
        tcache_entry *entries[TCACHE_MAX_BINS];
    } tcache_perthread_struct;
    """

    _fields_ = [
        ("num_slots", ctypes.c_uint16 * TCACHE_MAX_BINS),
        ("entries", c_pvoid * TCACHE_MAX_BINS),
    ]


class TcachePerthreadStruct(CStruct2GDB):
    """
    This class represents tcache_perthread_struct with interface compatible with `pwndbg.dbg_mod.Value`.
    """

    if GLIBC_VERSION >= (2, 42):
        _c_struct = c_tcache_perthread_struct_2_42
    elif GLIBC_VERSION >= (2, 30):
        _c_struct = c_tcache_perthread_struct_2_30
    else:
        _c_struct = c_tcache_perthread_struct_2_29
    sizeof = ctypes.sizeof(_c_struct)


class c_tcache_entry_2_28(Structure):
    """
    This class represents the tcache_entry struct for GLIBC < 2.29 as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.28/malloc/malloc.c#L2888

    typedef struct tcache_entry
    {
        struct tcache_entry *next;
    } tcache_entry;
    """

    _fields_ = [("next", c_pvoid)]


class c_tcache_entry_2_29(Structure):
    """
    This class represents the tcache_entry struct for GLIBC >= 2.29 as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.34/malloc/malloc.c#L3013

    typedef struct tcache_entry
    {
        struct tcache_entry *next;
        /* This field exists to detect double frees.  */
        uintptr_t key;
    } tcache_entry;
    """

    _fields_ = [("next", c_pvoid), ("key", c_pvoid)]


class TcacheEntry(CStruct2GDB):
    """
    This class represents the tcache_entry struct with interface compatible with `pwndbg.dbg_mod.Value`.
    """

    if GLIBC_VERSION >= (2, 29):
        _c_struct = c_tcache_entry_2_29
    else:
        _c_struct = c_tcache_entry_2_28
    sizeof = ctypes.sizeof(_c_struct)


class c_malloc_par_2_23(Structure):
    """
    This class represents the malloc_par struct for GLIBC < 2.24 as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.23/malloc/malloc.c#L1726

    struct malloc_par
    {
    /* Tunable parameters */
    unsigned long trim_threshold;
    INTERNAL_SIZE_T top_pad;
    INTERNAL_SIZE_T mmap_threshold;
    INTERNAL_SIZE_T arena_test;
    INTERNAL_SIZE_T arena_max;

    /* Memory map support */
    int n_mmaps;
    int n_mmaps_max;
    int max_n_mmaps;
    /* the mmap_threshold is dynamic, until the user sets
        it manually, at which point we need to disable any
        dynamic behavior. */
    int no_dyn_threshold;

    /* Statistics */
    INTERNAL_SIZE_T mmapped_mem;
    /*INTERNAL_SIZE_T  sbrked_mem;*/
    /*INTERNAL_SIZE_T  max_sbrked_mem;*/
    INTERNAL_SIZE_T max_mmapped_mem;
    INTERNAL_SIZE_T max_total_mem;  /* only kept for NO_THREADS */

    /* First address handed out by MORECORE/sbrk.  */
    char *sbrk_base;
    };
    """

    _fields_ = [
        ("trim_threshold", c_size_t),
        ("top_pad", c_size_t),
        ("mmap_threshold", c_size_t),
        ("arena_test", c_size_t),
        ("arena_max", c_size_t),
        ("n_mmaps", ctypes.c_int32),
        ("n_mmaps_max", ctypes.c_int32),
        ("max_n_mmaps", ctypes.c_int32),
        ("no_dyn_threshold", ctypes.c_int32),
        ("mmapped_mem", c_size_t),
        ("max_mmapped_mem", c_size_t),
        ("max_total_mem", c_size_t),
        ("sbrk_base", c_pvoid),
    ]


class c_malloc_par_2_12(Structure):
    """
    This class represents the malloc_par struct for GLIBC < 2.15 as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.12/malloc/malloc.c#L2402-L2433

    struct malloc_par {
    /* Tunable parameters */
    unsigned long    trim_threshold;
    INTERNAL_SIZE_T  top_pad;
    INTERNAL_SIZE_T  mmap_threshold;
    #ifdef PER_THREAD
    INTERNAL_SIZE_T  arena_test;
    INTERNAL_SIZE_T  arena_max;
    #endif

    /* Memory map support */
    int              n_mmaps;
    int              n_mmaps_max;
    int              max_n_mmaps;
    /* the mmap_threshold is dynamic, until the user sets
        it manually, at which point we need to disable any
        dynamic behavior. */
    int              no_dyn_threshold;

    /* Cache malloc_getpagesize */
    unsigned int     pagesize;

    /* Statistics */
    INTERNAL_SIZE_T  mmapped_mem;
    /*INTERNAL_SIZE_T  sbrked_mem;*/
    /*INTERNAL_SIZE_T  max_sbrked_mem;*/
    INTERNAL_SIZE_T  max_mmapped_mem;
    INTERNAL_SIZE_T  max_total_mem; /* only kept for NO_THREADS */

    /* First address handed out by MORECORE/sbrk.  */
    char*            sbrk_base;
    };
    """

    _fields_ = [
        ("trim_threshold", c_size_t),
        ("top_pad", c_size_t),
        ("mmap_threshold", c_size_t),
        ("arena_test", c_size_t),
        ("arena_max", c_size_t),
        ("n_mmaps", ctypes.c_int32),
        ("n_mmaps_max", ctypes.c_int32),
        ("max_n_mmaps", ctypes.c_int32),
        ("no_dyn_threshold", ctypes.c_int32),
        ("pagesize", c_size_t),
        ("mmapped_mem", c_size_t),
        ("max_mmapped_mem", c_size_t),
        ("max_total_mem", c_size_t),
        ("sbrk_base", c_pvoid),
    ]


class c_malloc_par_2_24(Structure):
    """
    This class represents the malloc_par struct for GLIBC >= 2.24 as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.25/malloc/malloc.c#L1690
    https://github.com/bminor/glibc/blob/glibc-2.24/malloc/malloc.c#L1719

    struct malloc_par
    {
        /* Tunable parameters */
        unsigned long trim_threshold;
        INTERNAL_SIZE_T top_pad;
        INTERNAL_SIZE_T mmap_threshold;
        INTERNAL_SIZE_T arena_test;
        INTERNAL_SIZE_T arena_max;

        /* Memory map support */
        int n_mmaps;
        int n_mmaps_max;
        int max_n_mmaps;
        /* the mmap_threshold is dynamic, until the user sets
            it manually, at which point we need to disable any
            dynamic behavior. */
        int no_dyn_threshold;

        /* Statistics */
        INTERNAL_SIZE_T mmapped_mem;
        INTERNAL_SIZE_T max_mmapped_mem;

        /* First address handed out by MORECORE/sbrk.  */
        char *sbrk_base;
    };
    """

    _fields_ = [
        ("trim_threshold", c_size_t),
        ("top_pad", c_size_t),
        ("mmap_threshold", c_size_t),
        ("arena_test", c_size_t),
        ("arena_max", c_size_t),
        ("n_mmaps", ctypes.c_int32),
        ("n_mmaps_max", ctypes.c_int32),
        ("max_n_mmaps", ctypes.c_int32),
        ("no_dyn_threshold", ctypes.c_int32),
        ("mmapped_mem", c_size_t),
        ("max_mmapped_mem", c_size_t),
        ("sbrk_base", c_pvoid),
    ]


class c_malloc_par_2_26(Structure):
    """
    This class represents the malloc_par struct for GLIBC >= 2.26 as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.34/malloc/malloc.c#L1875

    struct malloc_par
    {
        /* Tunable parameters */
        unsigned long trim_threshold;
        INTERNAL_SIZE_T top_pad;
        INTERNAL_SIZE_T mmap_threshold;
        INTERNAL_SIZE_T arena_test;
        INTERNAL_SIZE_T arena_max;

        /* Memory map support */
        int n_mmaps;
        int n_mmaps_max;
        int max_n_mmaps;
        /* the mmap_threshold is dynamic, until the user sets
            it manually, at which point we need to disable any
            dynamic behavior. */
        int no_dyn_threshold;

        /* Statistics */
        INTERNAL_SIZE_T mmapped_mem;
        INTERNAL_SIZE_T max_mmapped_mem;

        /* First address handed out by MORECORE/sbrk.  */
        char *sbrk_base;

    #if USE_TCACHE
        /* Maximum number of buckets to use.  */
        size_t tcache_bins;
        size_t tcache_max_bytes;
        /* Maximum number of chunks in each bucket.  */
        size_t tcache_count;
        /* Maximum number of chunks to remove from the unsorted list, which
            aren't used to prefill the cache.  */
        size_t tcache_unsorted_limit;
    #endif
    };
    """

    _fields_ = [
        ("trim_threshold", c_size_t),
        ("top_pad", c_size_t),
        ("mmap_threshold", c_size_t),
        ("arena_test", c_size_t),
        ("arena_max", c_size_t),
        ("n_mmaps", ctypes.c_int32),
        ("n_mmaps_max", ctypes.c_int32),
        ("max_n_mmaps", ctypes.c_int32),
        ("no_dyn_threshold", ctypes.c_int32),
        ("mmapped_mem", c_size_t),
        ("max_mmapped_mem", c_size_t),
        ("sbrk_base", c_pvoid),
        ("tcache_bins", c_size_t),
        ("tcache_max_bytes", c_size_t),
        ("tcache_count", c_size_t),
        ("tcache_unsorted_limit", c_size_t),
    ]


class c_malloc_par_2_35(Structure):
    """
    This class represents the malloc_par struct for GLIBC >= 2.35 as a ctypes struct.

    https://github.com/bminor/glibc/blob/glibc-2.35/malloc/malloc.c#L1874

    struct malloc_par
    {
        /* Tunable parameters */
        unsigned long trim_threshold;
        INTERNAL_SIZE_T top_pad;
        INTERNAL_SIZE_T mmap_threshold;
        INTERNAL_SIZE_T arena_test;
        INTERNAL_SIZE_T arena_max;

    #if HAVE_TUNABLES
        /* Transparent Large Page support.  */
        INTERNAL_SIZE_T thp_pagesize;
        /* A value different than 0 means to align mmap allocation to hp_pagesize
            add hp_flags on flags.  */
        INTERNAL_SIZE_T hp_pagesize;
        int hp_flags;
    #endif

        /* Memory map support */
        int n_mmaps;
        int n_mmaps_max;
        int max_n_mmaps;
        /* the mmap_threshold is dynamic, until the user sets
            it manually, at which point we need to disable any
            dynamic behavior. */
        int no_dyn_threshold;

        /* Statistics */
        INTERNAL_SIZE_T mmapped_mem;
        INTERNAL_SIZE_T max_mmapped_mem;

        /* First address handed out by MORECORE/sbrk.  */
        char *sbrk_base;

    #if USE_TCACHE
        /* Maximum number of buckets to use.  */
        size_t tcache_bins;
        size_t tcache_max_bytes;
        /* Maximum number of chunks in each bucket.  */
        size_t tcache_count;
        /* Maximum number of chunks to remove from the unsorted list, which
            aren't used to prefill the cache.  */
        size_t tcache_unsorted_limit;
    #endif
    };
    """

    _fields_ = [
        ("trim_threshold", c_size_t),
        ("top_pad", c_size_t),
        ("mmap_threshold", c_size_t),
        ("arena_test", c_size_t),
        ("arena_max", c_size_t),
        ("thp_pagesize", c_size_t),
        ("hp_pagesize", c_size_t),
        ("hp_flags", ctypes.c_int32),
        ("n_mmaps", ctypes.c_int32),
        ("n_mmaps_max", ctypes.c_int32),
        ("max_n_mmaps", ctypes.c_int32),
        ("no_dyn_threshold", ctypes.c_int32),
        ("mmapped_mem", c_size_t),
        ("max_mmapped_mem", c_size_t),
        ("sbrk_base", c_pvoid),
        ("tcache_bins", c_size_t),
        ("tcache_max_bytes", c_size_t),
        ("tcache_count", c_size_t),
        ("tcache_unsorted_limit", c_size_t),
    ]


class c_malloc_par_2_42(Structure):
    """
    This class represents the malloc_par struct for GLIBC >= 2.42 as a ctypes struct.

    https://elixir.bootlin.com/glibc/glibc-2.42/source/malloc/malloc.c#L1864

    struct malloc_par
    {
      /* Tunable parameters */
      unsigned long trim_threshold;
      INTERNAL_SIZE_T top_pad;
      INTERNAL_SIZE_T mmap_threshold;
      INTERNAL_SIZE_T arena_test;
      INTERNAL_SIZE_T arena_max;

      /* Transparent Large Page support.  */
      INTERNAL_SIZE_T thp_pagesize;
      /* A value different than 0 means to align mmap allocation to hp_pagesize
        add hp_flags on flags.  */
      INTERNAL_SIZE_T hp_pagesize;
      int hp_flags;

      /* Memory map support */
      int n_mmaps;
      int n_mmaps_max;
      int max_n_mmaps;
      /* the mmap_threshold is dynamic, until the user sets
        it manually, at which point we need to disable any
        dynamic behavior. */
      int no_dyn_threshold;

      /* Statistics */
      INTERNAL_SIZE_T mmapped_mem;
      INTERNAL_SIZE_T max_mmapped_mem;

      /* First address handed out by MORECORE/sbrk.  */
      char *sbrk_base;

    #if USE_TCACHE
      /* Maximum number of small buckets to use.  */
      size_t tcache_small_bins;
      size_t tcache_max_bytes;
      /* Maximum number of chunks in each bucket.  */
      size_t tcache_count;
      /* Maximum number of chunks to remove from the unsorted list, which
        aren't used to prefill the cache.  */
      size_t tcache_unsorted_limit;
    #endif
    };
    """

    _fields_ = [
        ("trim_threshold", c_size_t),
        ("top_pad", c_size_t),
        ("mmap_threshold", c_size_t),
        ("arena_test", c_size_t),
        ("arena_max", c_size_t),
        ("thp_pagesize", c_size_t),
        ("hp_pagesize", c_size_t),
        ("hp_flags", ctypes.c_int32),
        ("n_mmaps", ctypes.c_int32),
        ("n_mmaps_max", ctypes.c_int32),
        ("max_n_mmaps", ctypes.c_int32),
        ("no_dyn_threshold", ctypes.c_int32),
        ("mmapped_mem", c_size_t),
        ("max_mmapped_mem", c_size_t),
        ("sbrk_base", c_pvoid),
        ("tcache_small_bins", c_size_t),
        ("tcache_max_bytes", c_size_t),
        ("tcache_count", c_size_t),
        ("tcache_unsorted_limit", c_size_t),
    ]


class MallocPar(CStruct2GDB):
    """
    This class represents the malloc_par struct with interface compatible with `pwndbg.dbg_mod.Value`.
    """

    if GLIBC_VERSION >= (2, 42):
        _c_struct = c_malloc_par_2_42
    elif GLIBC_VERSION >= (2, 35):
        _c_struct = c_malloc_par_2_35
    elif GLIBC_VERSION >= (2, 26):
        _c_struct = c_malloc_par_2_26
    elif GLIBC_VERSION >= (2, 24):
        _c_struct = c_malloc_par_2_24
    elif GLIBC_VERSION >= (2, 15):
        _c_struct = c_malloc_par_2_23
    else:
        _c_struct = c_malloc_par_2_12
    sizeof = ctypes.sizeof(_c_struct)


# https://github.com/bminor/glibc/blob/glibc-2.37/malloc/malloc.c#L1911-L1926
# static struct malloc_par mp_ =
# {
#   .top_pad = DEFAULT_TOP_PAD,
#   .n_mmaps_max = DEFAULT_MMAP_MAX,
#   .mmap_threshold = DEFAULT_MMAP_THRESHOLD,
#   .trim_threshold = DEFAULT_TRIM_THRESHOLD,
# #define NARENAS_FROM_NCORES(n) ((n) * (sizeof (long) == 4 ? 2 : 8))
#   .arena_test = NARENAS_FROM_NCORES (1)
# #if USE_TCACHE
#   ,
#   .tcache_count = TCACHE_FILL_COUNT,
#   .tcache_bins = TCACHE_MAX_BINS,
#   .tcache_max_bytes = tidx2usize (TCACHE_MAX_BINS-1),
#   .tcache_unsorted_limit = 0 /* No limit.  */
# #endif
# };
DEFAULT_MP_ = MallocPar._c_struct()
DEFAULT_MP_.top_pad = DEFAULT_TOP_PAD
DEFAULT_MP_.n_mmaps_max = DEFAULT_MMAP_MAX
DEFAULT_MP_.mmap_threshold = DEFAULT_MMAP_THRESHOLD
DEFAULT_MP_.trim_threshold = DEFAULT_TRIM_THRESHOLD
DEFAULT_MP_.arena_test = 2 if pwndbg.aglib.arch.ptrsize == 4 else 8
if (MallocPar._c_struct != c_malloc_par_2_23) and (MallocPar._c_struct != c_malloc_par_2_12):
    # the only difference between 2.23 and the rest is the lack of tcache
    DEFAULT_MP_.tcache_count = TCACHE_FILL_COUNT
    if MallocPar._c_struct == c_malloc_par_2_42:
        DEFAULT_MP_.tcache_small_bins = TCACHE_SMALL_BINS
        DEFAULT_MP_.tcache_max_bytes = (
            MAX_TCACHE_SMALL_SIZE + SIZE_SZ + MALLOC_ALIGN_MASK
        ) & ~MALLOC_ALIGN_MASK + 1

    else:
        DEFAULT_MP_.tcache_bins = TCACHE_SMALL_BINS
        DEFAULT_MP_.tcache_max_bytes = MAX_TCACHE_SMALL_SIZE
if MallocPar._c_struct == c_malloc_par_2_12:
    DEFAULT_MP_.pagesize = DEFAULT_PAGE_SIZE



from __future__ import annotations

import os
import subprocess


def build_id() -> str:
    """
    Returns pwndbg commit id if git is available.
    """
    pwndbg_dir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
    # If we install pwndbg into site-packages, then `.pwndbg_root` is missing.
    if not os.path.exists(os.path.join(pwndbg_dir, ".pwndbg_root")):
        return ""

    try:
        git_path = os.path.join(pwndbg_dir, ".git")
        cmd = ["git", "--git-dir", git_path, "rev-parse", "--short", "HEAD"]

        commit_id = subprocess.check_output(cmd, stderr=subprocess.STDOUT)

        return "build: {}".format(commit_id.decode("utf-8").strip("\n"))

    except (OSError, subprocess.CalledProcessError):
        # OSError -> no git in $PATH
        # CalledProcessError -> git return code != 0
        return ""


__version__ = "2026.02.18"

b_id = build_id()

if b_id:
    __version__ += f" {b_id}"



from __future__ import annotations

from collections.abc import Generator

import pwndbg
import pwndbg.aglib.kernel.symbol
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.dbg_mod
import pwndbg.lib.cache
from pwndbg.aglib import kernel
from pwndbg.aglib.kernel.macros import compound_head
from pwndbg.aglib.kernel.macros import for_each_entry
from pwndbg.aglib.kernel.macros import swab


def caches() -> Generator[SlabCache, None, None]:
    recover_slab_typeinfo()
    slab_caches = pwndbg.aglib.kernel.slab_caches()
    if slab_caches is None:
        # Symbol not found
        return

    for slab_cache in for_each_entry(slab_caches, "struct kmem_cache", "list"):
        yield SlabCache(slab_cache)


def get_cache(target_name: str) -> SlabCache | None:
    slab_caches = pwndbg.aglib.kernel.slab_caches()
    if slab_caches is None:
        # Symbol not found
        return None

    for slab_cache in for_each_entry(slab_caches, "struct kmem_cache", "list"):
        if target_name == slab_cache["name"].string():
            return SlabCache(slab_cache)
    return None


def slab_struct_type() -> str:
    # In Linux kernel version 5.17 a slab struct was introduced instead of the previous page struct
    krelease = kernel.krelease()
    if krelease and krelease >= (5, 17):
        return "slab"
    return "page"


OO_SHIFT = 16
OO_MASK = (1 << OO_SHIFT) - 1


def oo_order(x: int) -> int:
    return int(x) >> OO_SHIFT


def oo_objects(x: int) -> int:
    return int(x) & OO_MASK


_flags = {
    "SLAB_DEBUG_FREE": 0x00000100,
    "SLAB_RED_ZONE": 0x00000400,
    "SLAB_POISON": 0x00000800,
    "SLAB_HWCACHE_ALIGN": 0x00002000,
    "SLAB_CACHE_DMA": 0x00004000,
    "SLAB_STORE_USER": 0x00010000,
    "SLAB_RECLAIM_ACCOUNT": 0x00020000,
    "SLAB_PANIC": 0x00040000,
    "SLAB_DESTROY_BY_RCU": 0x00080000,
    "SLAB_MEM_SPREAD": 0x00100000,
    "SLAB_TRACE": 0x00200000,
    "SLAB_DEBUG_OBJECTS": 0x00400000,
    "SLAB_NOLEAKTRACE": 0x00800000,
    "SLAB_NOTRACK": 0x01000000,
    "SLAB_FAILSLAB": 0x02000000,
}


def get_flags_list(flags: int) -> list[str]:
    return [flag_name for flag_name, mask in _flags.items() if flags & mask]


class Freelist:
    def __init__(self, start_addr: int, slab: Slab | None) -> None:
        self.start_addr = start_addr
        self.slab = slab
        if not self.slab:
            return
        self.offset = self.slab.slab_cache.offset
        self.random = self.slab.slab_cache.random
        self.cyclic = None

    def __iter__(self) -> Generator[int, None, None]:
        if not self.slab:
            return
        seen: set[int] = set()
        curr = None
        next = self.start_addr
        while next:
            if next in seen:
                self.cyclic = curr
                return
            if not pwndbg.aglib.memory.is_kernel(next + self.offset):
                break
            if next not in self.slab or not self.is_valid_obj(next):
                break
            curr = next
            next = self.find_next(curr)
            yield curr
            seen.add(curr)
        # reaching here means the freelist is not cyclic (prior to detections of other corruptions)
        self.cyclic = None

    def __int__(self) -> int:
        return self.start_addr

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def find_next(self, addr: int) -> int:
        # assumes addr is in this freelist -> assert(addr in self)
        # caller should assert this behaviour to avoid traversing the list unnecessarily
        if not self.slab:
            raise ValueError("slab freelist must belong to a slab")
        next = pwndbg.aglib.memory.read_pointer_width(addr + self.offset)
        if self.random:
            next ^= self.random ^ swab(addr + self.offset)
        return next

    def is_valid_obj(self, addr: int) -> bool:
        if not self.slab:
            return False
        diff = addr - self.slab.virt_address
        sz = self.slab.slab_cache.size
        return diff % sz == 0 and 0 <= (diff // sz) < self.slab.object_count


class SlabCache:
    def __init__(self, slab_cache: pwndbg.dbg_mod.Value) -> None:
        self._slab_cache = slab_cache

    @property
    def address(self) -> int:
        return int(self._slab_cache)

    @property
    def name(self) -> str:
        return self._slab_cache["name"].string()

    @property
    def offset(self) -> int:
        return int(self._slab_cache["offset"])

    @property
    def random(self) -> int:
        return (
            int(self._slab_cache["random"]) if "SLAB_FREELIST_HARDENED" in kernel.kconfig() else 0
        )

    @property
    def size(self) -> int:
        return int(self._slab_cache["size"])

    @property
    def slab_size(self) -> int:
        return kernel.page_size() << self.oo_order

    @property
    def object_size(self) -> int:
        return int(self._slab_cache["object_size"])

    @property
    def align(self) -> int:
        return int(self._slab_cache["align"])

    @property
    def flags(self) -> list[str]:
        return get_flags_list(int(self._slab_cache["flags"]))

    @property
    def cpu_cache(self) -> CpuCache | None:
        """returns cpu cache associated to current thread"""
        if not self._slab_cache.dereference().type.has_field("cpu_slab"):
            return None
        cpu = pwndbg.aglib.kernel.current_cpu()
        cpu_cache = kernel.per_cpu(self._slab_cache["cpu_slab"], cpu=cpu)
        return CpuCache(cpu_cache, self, cpu)

    @property
    def cpu_caches(self) -> Generator[CpuCache, None, None]:
        if not self._slab_cache.dereference().type.has_field("cpu_slab"):
            return
        """returns cpu caches for all cpus"""
        for cpu in range(kernel.nproc()):
            cpu_cache = kernel.per_cpu(self._slab_cache["cpu_slab"], cpu=cpu)
            yield CpuCache(cpu_cache, self, cpu)

    @property
    def node_caches(self) -> Generator[NodeCache, None, None]:
        """returns node caches for all NUMA nodes"""
        for node in range(kernel.num_numa_nodes()):
            yield NodeCache(self._slab_cache["node"][node], self, node)

    @property
    def cpu_partial(self) -> int | None:
        if not self._slab_cache.dereference().type.has_field("cpu_partial"):
            return None
        return int(self._slab_cache["cpu_partial"])

    @property
    def cpu_partial_slabs(self) -> int | None:
        if self._slab_cache.dereference().type.has_field(f"cpu_partial_{slab_struct_type()}s"):
            return int(self._slab_cache[f"cpu_partial_{slab_struct_type()}s"])
        return None

    @property
    def min_partial(self) -> int:
        return int(self._slab_cache["min_partial"])

    @property
    def inuse(self) -> int:
        # somewhat mirrors libslub's implementation
        # looks for per_cpu active lists and per_cpu and node partial lists
        # no good way to track full slabs unless CONFIG_SLUB_DEBUG is enabled
        #       which is typically not from what I have seen
        cnt = 0
        for cpu_cache in self.cpu_caches:
            if cpu_cache.active_slab is not None:
                cnt += cpu_cache.active_slab.inuse
            for partial_slab in cpu_cache.partial_slabs:
                cnt += partial_slab.inuse
        for node_cache in self.node_caches:
            for partial_slab in node_cache.partial_slabs:
                cnt += partial_slab.inuse
        return cnt

    @property
    def useroffset(self) -> int | None:
        if not self._slab_cache.dereference().type.has_field("useroffset"):
            return None
        return int(self._slab_cache["useroffset"])

    @property
    def usersize(self) -> int | None:
        if not self._slab_cache.dereference().type.has_field("usersize"):
            return None
        return int(self._slab_cache["usersize"])

    @property
    def __oo_x(self) -> int:
        return int(self._slab_cache["oo"]["x"])

    @property
    def oo_order(self) -> int:
        return oo_order(self.__oo_x)

    @property
    def oo_objects(self) -> int:
        return oo_objects(self.__oo_x)

    def find_containing_slab(self, address) -> Slab | None:
        for cpu_cache in self.cpu_caches:
            slab = cpu_cache.active_slab
            if slab is not None and address in slab:
                return slab
            for slab in cpu_cache.partial_slabs:
                if slab is not None and address in slab:
                    return slab
        for node_cache in self.node_caches:
            for slab in node_cache.partial_slabs:
                if slab is not None and address in slab:
                    return slab
        return None


class CpuCache:
    def __init__(self, cpu_cache: pwndbg.dbg_mod.Value, slab_cache: SlabCache, cpu: int) -> None:
        self._cpu_cache = cpu_cache
        self.slab_cache = slab_cache
        self.cpu = cpu

    @property
    def address(self) -> int:
        return int(self._cpu_cache)

    @property
    def freelist(self) -> Freelist:
        return Freelist(int(self._cpu_cache["freelist"]), self.active_slab)

    @property
    def active_slab(self) -> Slab | None:
        slab_key = slab_struct_type()
        _slab = self._cpu_cache[slab_key]
        if not int(_slab):
            return None
        return Slab(_slab.dereference(), cpu_cache=self, is_active=True)

    @property
    def partial_slabs(self) -> list[Slab]:
        partial_slabs = []
        if not self._cpu_cache.dereference().type.has_field("partial"):
            return []
        cur_slab = self._cpu_cache["partial"]
        cur_slab_int = int(cur_slab)
        while cur_slab_int:
            _slab = cur_slab.dereference()
            partial_slabs.append(Slab(_slab, cpu_cache=self))
            cur_slab = _slab["next"]
            cur_slab_int = int(cur_slab)
        return partial_slabs


class NodeCache:
    def __init__(self, node_cache: pwndbg.dbg_mod.Value, slab_cache: SlabCache, node: int) -> None:
        self._node_cache = node_cache
        self.slab_cache = slab_cache
        self.node = node

    @property
    def address(self) -> int:
        return int(self._node_cache)

    @property
    def partial_slabs(self) -> list[Slab]:
        ret = []
        for slab in for_each_entry(
            self._node_cache["partial"], f"struct {slab_struct_type()}", "slab_list"
        ):
            ret.append(Slab(slab.dereference(), node_cache=self))
        return ret

    @property
    def nr_partial(self) -> int:
        return int(self._node_cache["nr_partial"])

    @property
    def min_partial(self) -> int:
        return self.slab_cache.min_partial


class Slab:
    def __init__(
        self,
        slab: pwndbg.dbg_mod.Value,
        cpu_cache: CpuCache | None = None,
        node_cache: NodeCache | None = None,
        is_active: bool = False,
    ) -> None:
        self._slab = slab
        address = slab.address
        assert address
        self.slab_address = int(address)
        self.is_active = is_active
        if cpu_cache is not None:
            self.cpu_cache = cpu_cache
            self.is_cpu = True
            self.slab_cache = cpu_cache.slab_cache
        elif node_cache is not None:
            self.node_cache = node_cache
            self.is_cpu = False
            self.slab_cache = node_cache.slab_cache

    @property
    @pwndbg.lib.cache.cache_until("stop")
    def virt_address(self) -> int:
        if "CONFIG_SLAB_VIRTUAL" not in kernel.kconfig():
            return kernel.page_to_virt(self.slab_address)
        return kernel.slab_to_virt(self.slab_address)

    @property
    def object_count(self) -> int:
        return int(self._slab["objects"])

    @property
    def objects(self) -> Generator[int, None, None]:
        size = self.slab_cache.size
        start = self.virt_address
        end = start + self.object_count * size
        return (i for i in range(start, end, size))

    @property
    def frozen(self) -> int:
        return int(self._slab["frozen"])

    @property
    def inuse(self) -> int:
        inuse = int(self._slab["inuse"])
        if self.is_active:
            # only the cpu freelist is considered "inuse" similar to glibc's tcache
            inuse -= len(self.cpu_cache.freelist)
        return inuse

    @property
    def slabs(self) -> int:
        return int(self._slab[f"{slab_struct_type()}s"])

    @property
    def pobjects(self) -> int:
        if self.is_active:
            return 0
        if self._slab.type.has_field("pobjects"):
            return int(self._slab["pobjects"])
        # calculate approx obj count in half-full slabs (as done in kernel)
        # Note, this is a very bad approximation and could/should probably
        # be replaced by a more accurate method
        return (self.slabs * self.slab_cache.oo_objects) // 2

    @property
    def freelist(self) -> Freelist:
        return Freelist(int(self._slab["freelist"]), self)

    @property
    def free_objects(self) -> set[int]:
        result = set()
        for obj in self.freelist:
            result.add(obj)
        if self.is_active and self.cpu_cache.freelist:
            for obj in self.cpu_cache.freelist:
                result.add(obj)
        return result

    def __contains__(self, addr: int):
        return self.virt_address <= addr < self.virt_address + self.slab_cache.slab_size


def find_containing_slab_cache(addr: int) -> tuple[int | None, SlabCache | None]:
    recover_slab_typeinfo()  # throws a separate exception
    try:
        if "CONFIG_SLAB_VIRTUAL" not in kernel.kconfig():
            page = pwndbg.aglib.memory.get_typed_pointer_value(
                "struct page", kernel.virt_to_page(addr)
            )
            head_page = compound_head(page)
            address = head_page.address
            assert address
            base = kernel.page_to_virt(int(address))
            slab_type = pwndbg.aglib.typeinfo.load(f"struct {slab_struct_type()}")
            slab = head_page.cast(slab_type)
        else:
            _slab = kernel.virt_to_slab(addr)
            base = kernel.slab_to_virt(_slab)
            # mitigations are recent enough that the struct is expected to exist
            slab = pwndbg.aglib.memory.get_typed_pointer_value("struct slab", _slab)
        result = SlabCache(slab["slab_cache"])
        assert result.name  # make sure the result is sane
        return base, result
    except Exception:
        pass
    return None, None


#########################################
# structurs relevant to slab
#
#########################################


def kmem_cache_node_pad_sz(val: int) -> int | None:
    ptrsize = pwndbg.aglib.arch.ptrsize
    for j in range(8):
        nr_partial = pwndbg.aglib.memory.u32(val)
        next = pwndbg.aglib.memory.read_pointer_width(val + ptrsize)
        prev = pwndbg.aglib.memory.read_pointer_width(val + ptrsize * 2)
        val += ptrsize
        if (
            nr_partial < 0x20
            and pwndbg.aglib.memory.is_kernel(next)
            and pwndbg.aglib.memory.is_kernel(prev)
        ):
            return j * ptrsize
    return None


def get_kmem_cache():
    slab_caches = kernel.slab_caches()
    assert slab_caches, "can't find slab_caches"
    return int(slab_caches["prev"]) & ~0xFF


def kmem_cache_pad_sz() -> tuple[int, int]:
    # find the distance between the first kmem_cache's name and its first node cache
    # the name for the first kmem_cache (most likely) has the name "kmem_cache"
    # and the global var is also named "kmem_cache"
    kconfig = kernel.kconfig()
    name = "kmem_cache"
    name_off = None
    ptrsize = pwndbg.aglib.arch.ptrsize
    kmem_cache = get_kmem_cache()
    for i in range(0x20):
        val = pwndbg.aglib.memory.read_pointer_width(kmem_cache + i * ptrsize)
        if pwndbg.aglib.memory.string(val) == name.encode():
            name_off = i * ptrsize
            break
    assert name_off, "can't determine kmem_cache name offset"
    distance, node_cache_pad = None, None
    krelease = kernel.krelease()
    kasan_config_name = (
        "CONFIG_KASAN_GENERIC" if not krelease or krelease >= (6, 3) else "CONFIG_KASAN"
    )
    if (not krelease or krelease >= (6, 2)) and all(
        config not in kconfig for config in ("CONFIG_HARDENED_USERCOPY", kasan_config_name)
    ):
        if all(
            config not in kconfig
            for config in ("CONFIG_SYSFS", "CONFIG_SLAB_FREELIST_HARDENED", "CONFIG_NUMA")
        ):
            node_cache_pad = kmem_cache_node_pad_sz(
                kmem_cache + name_off + ptrsize * 3
            )  # name ptr + 2 list ptrs
            assert node_cache_pad, "can't find kmem_cache node"
            distance = ptrsize if "CONFIG_SLAB_FREELIST_RANDOM" in kconfig else 0
            return distance, node_cache_pad
        if "CONFIG_SLAB_FREELIST_RANDOM" in kconfig:
            for i in range(3, 0x20):
                ptr = kmem_cache + name_off + i * ptrsize
                val = pwndbg.aglib.memory.read_pointer_width(ptr)
                if pwndbg.aglib.memory.is_kernel(val) and all(
                    pwndbg.aglib.memory.uint(val + i * 4) < 0x10000 for i in range(10)
                ):  # checks the random_seq ptr for kmem_cache cache
                    _distance = (i + 1) * ptrsize
                    val = pwndbg.aglib.memory.read_pointer_width(kmem_cache + name_off + _distance)
                    node_cache_pad = kmem_cache_node_pad_sz(val)
                    if node_cache_pad is not None:
                        distance = _distance
                        break
            assert distance, "can't find kmem_cache node"
    if distance is None:
        for i in range(3, 0x20):
            ptr = kmem_cache + name_off + i * ptrsize
            val = pwndbg.aglib.memory.read_pointer_width(ptr - ptrsize)
            if pwndbg.aglib.memory.peek(val) is not None:
                continue
            val = pwndbg.aglib.memory.read_pointer_width(ptr)
            if pwndbg.aglib.memory.peek(val) is None:
                continue
            node_cache_pad = kmem_cache_node_pad_sz(val)
            if node_cache_pad is not None:
                distance = i * ptrsize
                break
    assert distance is not None and node_cache_pad is not None, "can't find kmem_cache node"
    distance -= ptrsize * 3  # the name ptr + list_head
    for config in ("CONFIG_SLAB_FREELIST_HARDENED", "CONFIG_NUMA", "CONFIG_SLAB_FREELIST_RANDOM"):
        if config in kconfig:
            distance -= ptrsize
    if kasan_config_name in kconfig and krelease:  # kasan
        if krelease >= (6, 3) or krelease < (6, 1) or "CONFIG_KASAN_GENERIC" in kernel.kconfig():
            distance -= pwndbg.aglib.typeinfo.uint.sizeof * 2
        if (5, 12) <= krelease and krelease < (6, 3):
            distance -= ptrsize
    if "CONFIG_HARDENED_USERCOPY" in kconfig or (krelease and krelease < (6, 2)):
        distance -= pwndbg.aglib.typeinfo.uint.sizeof * 2
    assert distance < 0x1000, "cannot find kmem_cache padding size"
    return distance, node_cache_pad


def get_kmem_cache_padding_sz() -> int:
    ptrsize = pwndbg.aglib.arch.ptrsize
    off = ptrsize  # __page_flags
    if "CONFIG_SLAB_VIRTUAL" in kernel.kconfig():
        off = None
        # per cpu cache always exists because CONFIG_SLAB_VIRTUAL -> !CONFIG_SLUB_TINY
        kmem_cache_addr = get_kmem_cache()
        kmem_cache_cpu = pwndbg.aglib.memory.read_pointer_width(kmem_cache_addr)
        slabs = []
        for i in range(pwndbg.aglib.kernel.nproc()):
            _cache_cpu = int(pwndbg.aglib.kernel.per_cpu(kmem_cache_cpu, i))
            ptr = pwndbg.aglib.memory.read_pointer_width(
                _cache_cpu + ptrsize * 2
            )  # struct slab *slab
            if pwndbg.aglib.memory.is_kernel(ptr):
                slabs.append(ptr)
            if "CONFIG_SLUB_CPU_PARTIAL" in kernel.kconfig():
                ptr = pwndbg.aglib.memory.read_pointer_width(_cache_cpu + ptrsize * 3)  # partial
                if pwndbg.aglib.memory.is_kernel(ptr):
                    slabs.append(ptr)
        for slab in slabs:
            for i in range(1, 0x10):  # skip the first slab ptr, find the ptr to kmem_cache
                # find kmem_cache ptr which should be in slab virtual range (allocated with SLUB)
                addr = slab + i * ptrsize
                if not pwndbg.aglib.memory.is_kernel(addr):
                    continue
                val = pwndbg.aglib.memory.read_pointer_width(addr)
                if kernel.slab_virtual() < val and pwndbg.aglib.memory.is_kernel(val):
                    off = ptrsize * i
                    break
            if off is not None:
                break
        if off is None:
            off = 7 * ptrsize  # default value for mitigation-6.12
    return off


def kmem_cache_structs(node_cache_pad: int) -> str:
    if pwndbg.aglib.kernel.symbol.kversion_cint() is None:
        return ""
    result = f"#define KVERSION {pwndbg.aglib.kernel.symbol.kversion_cint()}\n"
    if "CONFIG_SLUB_CPU_PARTIAL" in pwndbg.aglib.kernel.kconfig():
        result += "#define CONFIG_SLUB_CPU_PARTIAL\n"
    result += f"""
    struct kmem_cache_node {{
        char _pad[{node_cache_pad}];
        unsigned long nr_partial;
        struct list_head partial;
    }};
    """
    result += f"#define PADDING {get_kmem_cache_padding_sz()}\n"
    result += """
    struct kasan_cache {
#if !((KERNEL_VERSION(6, 1, 0) <= KVERSION && KVERSION < KERNEL_VERSION(6, 3, 0)))
        int alloc_meta_offset;
        int free_meta_offset;
#elif defined(CONFIG_KASAN_GENERIC)
        int alloc_meta_offset;
        int free_meta_offset;
#endif
#if KERNEL_VERSION(5, 12, 0) <= KVERSION && KVERSION < KERNEL_VERSION(6, 3, 0)
        bool is_kmalloc;
#endif
    };
    struct kmem_cache_order_objects {
        unsigned int x;
    };
    struct reciprocal_value {
        u32 m;
        u8 sh1, sh2;
    };
    typedef unsigned int gfp_t;
    typedef unsigned int slab_flags_t;
#if KVERSION >= KERNEL_VERSION(5, 17, 0)
    struct slab {
        char pad[PADDING];
#if KVERSION >= KERNEL_VERSION(6, 2, 0)
        struct kmem_cache *slab_cache;
#endif
        union {
            struct list_head slab_list;
#ifdef CONFIG_SLUB_CPU_PARTIAL
            struct {
                struct slab *next;
                int slabs;	/* Nr of slabs left */
            };
#endif
        };
#if KVERSION < KERNEL_VERSION(6, 2, 0)
        struct kmem_cache *slab_cache;
#endif
        void *freelist;		/* first free object */
        union {
            unsigned long counters;
            struct {
                unsigned inuse:16;
                unsigned objects:15;
                unsigned frozen:1;
            };
        };
        // rcu_head in later versions is not important for our purposes
#ifndef CONFIG_SLAB_VIRTUAL
        unsigned int __page_type;
        atomic_t __page_refcount;
#endif
        /* memcg data unused in pwndbg */
    };
#endif
    """
    result += f"""
    struct kmem_cache_cpu {{
        void **freelist;	/* Pointer to next available object */
        unsigned long tid;	/* Globally unique transaction id */
        struct {slab_struct_type()} *{slab_struct_type()};	/* The slab from which we are allocating */
#ifdef CONFIG_SLUB_CPU_PARTIAL
        struct {slab_struct_type()} *partial;	/* Partially allocated frozen slabs */
#endif
        /* irrelevant fields*/
    }};
    """
    if "CONFIG_SLAB_VIRTUAL" in kernel.kconfig():
        # TODO: the size of freed_slabs_lock is inaccurate but I'm not sure how to handle it
        result += """
        struct kmem_cache_virtual {
            int freed_slabs_lock; // spinlock_t -> this struct is complex cuz its config dep
            struct list_head freed_slabs;
            struct list_head freed_slabs_min;
            unsigned long nr_freed_pages;
        };
        """
    return result


@pwndbg.aglib.kernel.typeinfo_recovery("struct kmem_cache", requires_kversion=True)
def recover_slab_typeinfo() -> str:
    kconfig = pwndbg.aglib.kernel.kconfig()
    defs = []
    configs = (
        "CONFIG_SLUB_TINY",
        "CONFIG_SLUB_CPU_PARTIAL",
        "CONFIG_SLAB_FREELIST_HARDENED",
        "CONFIG_NUMA",
        "CONFIG_SLAB_FREELIST_RANDOM",
        "CONFIG_KASAN_GENERIC",
        "CONFIG_HARDENED_USERCOPY",
        "CONFIG_KASAN",
        "CONFIG_LOCKDEP",
        "CONFIG_SLAB_VIRTUAL",
    )
    for config in configs:
        if config in kconfig:
            defs.append(config)
    sz, node_cache_pad = kmem_cache_pad_sz()
    result = f"#define KVERSION {pwndbg.aglib.kernel.symbol.kversion_cint()}\n"
    result += "\n".join(f"#define {s}" for s in defs)
    result += pwndbg.aglib.kernel.symbol.COMMON_TYPES
    result += kmem_cache_structs(node_cache_pad)
    # this is the kmem_cache SLUB representation for all 5.x and 6.x
    result += f"""
    struct kmem_cache {{
#if !defined(CONFIG_SLUB_TINY) || KVERSION < KERNEL_VERSION(6, 2, 0)
        struct kmem_cache_cpu *cpu_slab;
#endif
#if KVERSION >= KERNEL_VERSION(6, 18, 0)
#if !defined(CONFIG_SLUB_TINY) && defined(CONFIG_LOCKDEP)
        char lock_key[POINTER_SIZE * 2]; // sizeof(hash_entry)
#endif
        void *cpu_sheaves;
#endif
        /* Used for retrieving partial slabs, etc. */
        slab_flags_t flags;
        unsigned long min_partial;
        unsigned int size;		/* Object size including metadata */
        unsigned int object_size;	/* Object size without metadata */
#if KVERSION >= KERNEL_VERSION(5, 9, 0)
        struct reciprocal_value reciprocal_size;
#endif
        unsigned int offset;		/* Free pointer offset */
#ifdef CONFIG_SLUB_CPU_PARTIAL
        /* Number of per cpu partial objects to keep around */
        unsigned int cpu_partial;
#if KVERSION >= KERNEL_VERSION(5, 16, 0)
        /* Number of per cpu partial slabs to keep around */
        unsigned int cpu_partial_{slab_struct_type()}s;
#endif
#endif
#if KVERSION >= KERNEL_VERSION(6, 18, 0)
        unsigned int sheaf_capacity;
#endif
        struct kmem_cache_order_objects oo;
        /* Allocation and freeing of slabs */
        struct kmem_cache_order_objects min;
#if KVERSION < KERNEL_VERSION(5, 19, 0)
        struct kmem_cache_order_objects max;
#endif
#ifdef CONFIG_SLAB_VIRTUAL
        struct kmem_cache_virtual virtual;
#endif
        gfp_t allocflags;		/* gfp flags to use on each alloc */
        int refcount;			/* Refcount for slab cache destroy */
        void *ctor;	            /* Object constructor -- ignoring possible args */
        unsigned int inuse;		/* Offset to metadata */
        unsigned int align;		/* Alignment */
        unsigned int red_left_pad;	/* Left redzone padding size */
        const char *name;		/* Name (only for display!) */
        struct list_head list;		/* List of slab caches */

        char _pad1[{sz}]; // collapse the struct(s) that are version dependent and complex
#ifdef CONFIG_SLAB_FREELIST_HARDENED
        unsigned long random;
#endif
#ifdef CONFIG_NUMA
        unsigned int remote_node_defrag_ratio;
#endif
#ifdef CONFIG_SLAB_FREELIST_RANDOM
        unsigned int *random_seq;
#endif
#if (KVERSION >= KERNEL_VERSION(6, 3, 0) && defined(CONFIG_KASAN_GENERIC) || (KVERSION < KERNEL_VERSION(6, 3, 0) && defined(CONFIG_KASAN)))
        struct kasan_cache kasan_info;
#endif
#if KVERSION < KERNEL_VERSION(6, 2, 0) || defined(CONFIG_HARDENED_USERCOPY)
        unsigned int useroffset;	/* Usercopy region offset */
        unsigned int usersize;		/* Usercopy region size */
#endif
        // ensure it has at least num_numa_nodes, sufficient for us
        struct kmem_cache_node *node[{kernel.num_numa_nodes()}];
    }};
    """
    return result



"""
Retrieve files from the debuggee's filesystem.  Useful when
debugging a remote process over SSH or similar, where e.g.
/proc/FOO/maps is needed from the remote system.
"""

from __future__ import annotations

import errno
import os
import shutil
import tempfile
from collections.abc import Iterator

import pwndbg.aglib.proc
import pwndbg.aglib.qemu
import pwndbg.aglib.remote
import pwndbg.dbg_mod
import pwndbg.lib.cache

_remote_files_dir = None


def reset_remote_files() -> None:
    global _remote_files_dir

    if _remote_files_dir is not None:
        shutil.rmtree(_remote_files_dir)
        _remote_files_dir = None


def remote_files_dir():
    global _remote_files_dir

    if _remote_files_dir is None:
        _remote_files_dir = tempfile.mkdtemp()

    return _remote_files_dir


def get_proc_exe_file() -> str:
    """
    Returns the local path to the debugged file name.
    """
    return get_file(pwndbg.aglib.proc.exe(), try_local_path=True)


@pwndbg.lib.cache.cache_until("start")
def can_download_remote_file() -> bool:
    if not pwndbg.aglib.remote.is_remote():
        return False
    if pwndbg.aglib.qemu.is_qemu_kernel():
        return False

    # Some[1] gdb servers don't implement vFile packets.
    # [1] - qemu-user <8.1
    # [1] - Rosetta2
    # [1] - maybe embedded probe like: `Black Magic Probe V2.3`
    # WTF: There is no indication in `qSupported` when `vFile` packets are supported
    # Probe and check what it returns
    try:
        vfile_open("", 0, 0)
    except OSError:
        return True
    except NotImplementedError:
        return False

    return False


def get_file(path: str, try_local_path: bool = False) -> str:
    """
    Downloads the specified file from the system where the current process is
    being debugged.

    If the `path` is prefixed with "target:" the prefix is stripped
    (to support remote target paths properly).

    If the `try_local_path` is set to `True` and the `path` exists locally and "target:" prefix is not present, it will return the local path instead of downloading the file.

    Returns:
        The local path to the file
    """
    has_target_prefix = path.startswith("target:")
    if has_target_prefix:
        path = path[7:]  # len('target:') == 7

    local_path = path
    if not pwndbg.aglib.remote.is_remote() or pwndbg.aglib.qemu.is_qemu_kernel():
        if not os.path.exists(local_path):
            raise OSError(f"File '{local_path}' does not exist", errno.ENOENT)

        return local_path

    if try_local_path and not has_target_prefix and os.path.exists(local_path):
        return local_path

    if can_download_remote_file():
        local_path = tempfile.mktemp(dir=remote_files_dir())
        try:
            pwndbg.dbg.selected_inferior().download_remote_file(path, local_path)
        except pwndbg.dbg_mod.Error as e:
            # This module originally raised this as an OSError.
            raise OSError(e)
    else:
        raise OSError(f"get_file('{local_path}') is not supported for your target", errno.ENODEV)

    return local_path


def get(path: str) -> bytes:
    """
    Retrieves the contents of the specified file on the system
    where the current process is being debugged.

    Returns:
        A byte array, or None.
    """
    local_path = get_file(path)

    try:
        with open(local_path, "rb") as f:
            return f.read()
    except Exception:
        return b""


def readlink(path: str) -> str:
    """readlink(path) -> str

    Read the link specified by 'path' on the system being debugged.

    Handles local, qemu-usermode, and remote debugging cases.
    """
    if pwndbg.aglib.remote.is_remote():
        # FIXME: implement `get_sysroot` when remote debugging,
        #  logic should be same as we do in `get_file`
        if can_download_remote_file():
            try:
                return vfile_readlink(path).decode("utf-8")
            except Exception:
                return ""
        return ""

    try:
        return os.readlink(path)
    except Exception:
        return ""


@pwndbg.lib.cache.cache_until("start")
def is_vfile_qemu_user_bug() -> bool:
    # This is a BUG[1] in the gdbstub of QEMU user mode. It should return data encoded in hexadecimal,
    # but instead, it returns the data as a decimal integer (%d).
    # [1] https://github.com/qemu/qemu/blob/b14d0649628cbe88ac0ef35fcf58cd1fc22735b8/gdbstub/user-target.c#L322
    if not pwndbg.aglib.qemu.is_qemu_usermode():
        return False

    # The bug was fixed in QEMU version v10.0.0-rc0 but the qGDBServerVersion packet wasn't added until v10.1.0-rc0
    # Therefore, check the simple case of whether the QEMU version >= v10.1.0
    # and fallback to the old check method otherwise to handle the gap between v10.0.0 and v10.1.0
    version = pwndbg.aglib.qemu.qemu_gdbserver_version()
    if version is not None and version >= (10, 1, 0):
        return False

    # On a bugged QEMU version, the response is `F-1,36`
    # On a fixed QEMU version, the response is `F-1,24`
    # This performs the syscall: `openat(0, "/\01*256", O_RDONLY|0x20) = -1 ENAMETOOLONG (File name too long)`
    response = pwndbg.dbg.selected_inferior().send_remote("vFile:open:2f" + ("01" * 256))
    return response == b"F-1,36"


def _vfile_check_response(response: bytes):
    if len(response) == 0:
        raise NotImplementedError("Not supported")
    # F result [,errno] [;attachment]; strip attachment first, skip if no errno
    result_errno = response.split(b";", 1)[0]
    if result_errno.startswith(b"F") and b"," in result_errno:
        result, errno = result_errno[1:].split(b",", 1)  # Skip "F"
        if int(result) != -1:
            raise NotImplementedError(f"Unknown response status {result!r}")
        err = int(errno, 10 if is_vfile_qemu_user_bug() else 16)
        raise OSError(err, "Error")


def vfile_readlink(pathname: str | bytes) -> bytes:
    """
    Reads the target of a symbolic link on the remote system.

    :param pathname: The path to the symbolic link (string).
    :param buffer_size: The size of the buffer to read into (integer).
    :return: The target of the symbolic link as a string.
    """
    if isinstance(pathname, str):
        pathname = pathname.encode("utf-8")
    encoded_pathname = pathname.hex()

    packet = f"vFile:readlink:{encoded_pathname}"
    response = pwndbg.dbg.selected_inferior().send_remote(packet)
    _vfile_check_response(response)

    parts = response[1:].split(b";", 1)
    # bytes_read = int(parts[0], 16)
    target = gdb_memtox_inverse(parts[1])
    return target


def vfile_readfile(filename: str, chunk_size=1000) -> Iterator[bytes]:
    """
    Reads the entire content of a file on the remote system.

    :param filename: The path to the file (string).
    :param chunk_size: The number of bytes to read in each iteration (integer).
    :return: The complete content of the file as bytes.
    """
    fd = None
    try:
        # 0 = readonly
        fd = vfile_open(filename, 0, 0)
        offset = 0

        while True:
            bytes_read, data = vfile_pread(fd, chunk_size, offset)
            if bytes_read == 0:
                break
            yield data
            offset += bytes_read
    finally:
        if fd is not None:
            vfile_close(fd)


def vfile_open(filename: str, flags: int, mode: int) -> int:
    """
    Opens a file on the remote system and returns the file descriptor.

    :param filename: The path to the file (string).
    :param flags: Flags passed to the open call (integer, base 16).
        These correspond to the constant values in the enum `OpenOptions` from LLDB’s `File.h`,
        not the traditional `open(2)` flags.
    :param mode: Mode bits for the file (integer, base 16).
    :return: File descriptor (integer), or raises an exception if an error occurs.
    """
    encoded_filename = filename.encode("utf-8").hex()
    packet = f"vFile:open:{encoded_filename},{flags:08X},{mode:08X}"
    response = pwndbg.dbg.selected_inferior().send_remote(packet)
    _vfile_check_response(response)

    file_descriptor = int(response[1:].decode(), 10 if is_vfile_qemu_user_bug() else 16)
    return file_descriptor


# GDB defines the encoding of binary data transferred through commands here:
# - https://sourceware.org/gdb/current/onlinedocs/gdb.html/Overview.html#:~:text=The%20binary%20data%20representation%20uses%207d
# This function unescapes these character
# Example of QEMU function that does this escaping:
# https://github.com/qemu/qemu/blob/de278e54aefed143526174335f8286f7437d20be/gdbstub/gdbstub.c#L184
def gdb_memtox_inverse(data: bytes) -> bytes:
    buffer = bytearray()

    i = 0
    while i < len(data):
        b = data[i]
        if b == 125:  # == ord("}"):
            buffer.append(data[i + 1] ^ 0x20)
            i += 1
        else:
            buffer.append(b)
        i += 1

    return buffer


def vfile_pread(fd: int, size: int, offset: int) -> tuple[int, bytes]:
    """
    Reads data from a file descriptor.

    :param fd: File descriptor (integer).
    :param size: Number of bytes to read (integer, base 16).
    :param offset: Offset in the file to start reading from (integer, base 16).
    :return: Tuple of (bytes_read, data) where bytes_read is an integer and data is the binary data.
    """
    packet = f"vFile:pread:{fd:X},{size:X},{offset:X}"
    response = pwndbg.dbg.selected_inferior().send_remote(packet)
    _vfile_check_response(response)

    parts = response[1:].split(b";", 1)
    bytes_read = int(parts[0].decode(), 16)
    # We have to decode the data, because some bytes -  #,$,*,} - have special encodings
    data = gdb_memtox_inverse(parts[1])

    return bytes_read, data


def vfile_close(fd):
    """
    Closes a previously opened file descriptor.

    :param fd: File descriptor (integer).
    :return: None, or raises an exception if an error occurs.
    """
    packet = f"vFile:close:{fd:X}"
    response = pwndbg.dbg.selected_inferior().send_remote(packet)
    _vfile_check_response(response)
    return None



"""
Displays information about kernel tasks. This command iterates through the kernel's task list
and prints details about each task, including its address, PID, user space status, CPU, UID, GID, and name.
"""

from __future__ import annotations

import argparse

import pwndbg.aglib.kernel
import pwndbg.aglib.symbol
import pwndbg.color as color
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.dbg_mod
import pwndbg.lib
import pwndbg.lib.cache
from pwndbg.aglib.kernel.macros import for_each_entry
from pwndbg.lib.exception import IndentContextManager

parser = argparse.ArgumentParser(description="Displays information about kernel tasks.")
parser.add_argument("task_name", nargs="?", type=str, help="A task name to search for")

indent = IndentContextManager()


class Kthread:
    def __init__(self, thread: pwndbg.dbg_mod.Value):
        self.thread = thread
        self.name = thread["comm"].string()
        self.pid = int(thread["pid"])
        self.has_user_page = int(thread["mm"]) != 0
        krelease = pwndbg.aglib.kernel.krelease()
        if krelease is None or "CONFIG_THREAD_INFO_IN_TASK" not in pwndbg.aglib.kernel.kconfig():
            self.cpu = "-"
        elif krelease < (5, 16):
            self.cpu = int(thread["cpu"])
        else:
            self.cpu = int(thread["thread_info"]["cpu"])
        self.uid = int(thread["real_cred"]["uid"]["val"])
        self.gid = int(thread["real_cred"]["gid"]["val"])

    @pwndbg.lib.cache.cache_until("stop")
    def files(self):
        fdt = self.thread["files"]["fdt"]
        fds = fdt["fd"]
        files = []
        for i in range(int(fdt["max_fds"])):
            file = fds[i]
            addr = int(file)
            if addr == 0:
                continue
            files.append((i, file))
        return tuple(files)

    @property
    def mm(self):
        mm = self.thread["mm"]
        if int(mm) != 0:
            return mm
        # for anonymous tasks
        mm = self.thread["active_mm"]
        if int(mm) != 0:
            return mm
        return None

    def __str__(self):
        thread = color.blue(hex(int(self.thread)))
        prefix = f"[pid {self.pid}]"
        desc = " "
        prefix = color.blue(f"{prefix:<9}") + f"task @ {thread}: {self.name:<16}"
        user = ", has user pages" if self.has_user_page else ""
        desc = color.red(f"cpu #{self.cpu} (uid: {self.uid}, gid: {self.gid}{user})")
        return f"{prefix} {desc}"


class Ktask:
    def __init__(self, task: pwndbg.dbg_mod.Value):
        self.task = task
        threads = []
        signal = task["signal"]
        # Iterate through all threads in the task_struct's thread list.
        for thread in for_each_entry(signal["thread_head"], "struct task_struct", "thread_node"):
            kthread = Kthread(thread)
            threads.append(kthread)
        self.threads = threads


@pwndbg.lib.cache.cache_until("stop")
def get_ktasks() -> tuple[Ktask, ...]:
    tasks = []
    # Look up the init_task symbol, which is the first task in the kernel's task list.
    init_task = pwndbg.aglib.symbol.lookup_symbol("init_task")
    if init_task is None:
        print(
            "The init_task symbol was not found. This may indicate that the symbol is not available in the current build."
        )
        return None

    try:
        tasks.append(Ktask(init_task))
        # The task list is implemented a circular doubly linked list, so we traverse starting from init_task.
        for task in for_each_entry(init_task["tasks"], "struct task_struct", "tasks"):
            ktask = Ktask(task)
            tasks.append(ktask)
    except pwndbg.dbg_mod.Error as e:
        print(message.error(f"ERROR: {e}"))
        return None
    return tuple(tasks)


@pwndbg.commands.Command(parser, category=pwndbg.commands.CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWhenPagingEnabled
@pwndbg.commands.OnlyWithKernelDebugInfo
def ktask(task_name=None) -> None:
    threads = []
    for task in get_ktasks():
        for thread in task.threads:
            if task_name is not None and task_name not in thread.name:
                continue
            threads.append(thread)
    for thread in threads:
        indent.print(thread)



from __future__ import annotations

import bisect

import pwndbg
import pwndbg.aglib.memory
import pwndbg.aglib.qemu
import pwndbg.aglib.stack
import pwndbg.color.message as message
import pwndbg.lib.cache
import pwndbg.lib.config
import pwndbg.lib.memory
from pwndbg.dbg_mod import EventType

# List of manually-explored pages which were discovered
# by analyzing the stack or register context.
explored_pages: list[pwndbg.lib.memory.Page] = []

# List of custom pages that can be managed manually by vmmap_* commands family
custom_pages: list[pwndbg.lib.memory.Page] = []

auto_explore = pwndbg.config.add_param(
    "auto-explore-pages",
    "warn",
    "whether to try to infer page permissions when memory maps are missing",
    param_class=pwndbg.lib.config.PARAM_ENUM,
    enum_sequence=["yes", "warn", "no"],
    help_docstring="""
This command can cause errors.
""",
)


_warn_cache: set[int] = set()


@pwndbg.dbg.event_handler(EventType.NEW_MODULE)
def clear_warn_cache():
    _warn_cache.clear()


def get_custom_pages() -> tuple[pwndbg.lib.memory.Page, ...]:
    """
    Returns a tuple of `Page` objects representing the memory mappings of the
    target, sorted by virtual address ascending.
    """
    pages: list[pwndbg.lib.memory.Page] = [*explored_pages, *custom_pages]
    pages.sort()
    return tuple(pages)


def add_custom_page(page: pwndbg.lib.memory.Page) -> None:
    bisect.insort(custom_pages, page)

    # Reset all the cache
    # We can not reset get() only, since the result may be used by others.
    # TODO: avoid flush all caches
    pwndbg.lib.cache.clear_caches()


def clear_custom_page() -> None:
    while custom_pages:
        custom_pages.pop()

    # Reset all the cache
    # We can not reset get() only, since the result may be used by others.
    # TODO: avoid flush all caches
    pwndbg.lib.cache.clear_caches()


def explore(address_maybe: int) -> pwndbg.lib.memory.Page | None:
    """
    Given a potential address, check to see what permissions it has.

    Returns:
        Page object

    Note:
        Adds the Page object to a persistent list of pages which are
        only reset when the process dies. This means pages which are
        added this way will not be removed when unmapped.

        Also assumes the entire contiguous section has the same permission.
    """
    if not pwndbg.aglib.memory.is_pagefault_supported():
        return None

    if auto_explore.value == "warn":
        page_start = pwndbg.lib.memory.page_align(address_maybe)
        if page_start not in _warn_cache:
            _warn_cache.add(page_start)
            is_readable_addr = pwndbg.aglib.memory.peek(page_start)
            if is_readable_addr:
                print(
                    message.warn(
                        f"Warning: Avoided exploring possible address {address_maybe:#x}.\n"
                        f"You can explicitly explore it with `vmmap-explore {page_start:#x}` or disable automatic exploration with `set auto-explore-pages no`."
                    )
                )
                if pwndbg.aglib.qemu.is_qemu_kernel() and pwndbg.aglib.memory.is_kernel(page_start):
                    print(message.error("Likely a pagescan bug, please report."))

        return None
    if auto_explore.value == "no":
        return None

    address_maybe = pwndbg.lib.memory.page_align(address_maybe)
    flags = get_memory_flags(address_maybe)
    if flags is None:
        return None

    page = find_boundaries(address_maybe)
    page.objfile = f"<explored_{address_maybe >> 12:05x}>"
    page.flags = flags

    explored_pages.append(page)

    # Reset all the cache
    # We can not reset get() only, since the result may be used by others.
    # TODO: avoid flush all caches
    pwndbg.lib.cache.clear_caches()

    return page


def get_memory_flags(address_maybe: int) -> int | None:
    flags = 4 if pwndbg.aglib.memory.peek(address_maybe) else 0
    if not flags:
        return None

    if pwndbg.aglib.memory.poke(address_maybe):
        flags |= 2

    # It's really hard to check for executability, so we just make some guesses:
    # 1. If it's in the same page as the instruction pointer, assume it's executable.
    # 2. If it's in the same page as the stack pointer, try to check the NX bit.
    # 3. Otherwise, just say it's not executable.
    if address_maybe == pwndbg.lib.memory.page_align(pwndbg.aglib.regs.pc):
        flags |= 1
    # TODO: could maybe make this check look at the stacks in pwndbg.aglib.stack.get() but that might have issues
    elif (
        address_maybe == pwndbg.lib.memory.page_align(pwndbg.aglib.regs.sp)
        and pwndbg.aglib.stack.is_executable()
    ):
        flags |= 1

    return flags


def find_boundaries(addr: int, name: str = "", min: int = 0) -> pwndbg.lib.memory.Page:
    """
    Given a single address, find all contiguous pages
    which are mapped.
    """
    start = pwndbg.aglib.memory.find_lower_boundary(addr)
    end = pwndbg.aglib.memory.find_upper_boundary(addr)

    start = max(start, min)
    end = max(end, min)

    return pwndbg.lib.memory.Page(start, end - start, 4, 0, pwndbg.aglib.arch.ptrsize, name)



from __future__ import annotations

import argparse

import pwndbg.aglib.kernel
import pwndbg.aglib.memory
import pwndbg.chain
import pwndbg.color as color
import pwndbg.color.context as ctx_color
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.commands.ktask
from pwndbg.lib.exception import IndentContextManager
from pwndbg.lib.regs import BitFlags

indent = IndentContextManager()

fmode_flags = BitFlags([("R", 0), ("W", 1), ("X", 5)])
KCURRENT_PID = None
KCURRENT_PGD = None

parser = argparse.ArgumentParser(
    description="Displays information about fds accessible by a kernel task."
)
parser.add_argument("pid", nargs="?", type=int, help="")
parser.add_argument("--fd", nargs="?", type=int, help="")


@pwndbg.commands.Command(parser, category=pwndbg.commands.CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWhenPagingEnabled
@pwndbg.commands.OnlyWithKernelDebugInfo
def kfile(pid=None, fd=None):
    if pid is None:
        if KCURRENT_PID is None:
            kcurrent(None, set_pid=True, verbose=False)
        pid = KCURRENT_PID
    if pid is None:
        print(message.warn("no pid specified (either specify pid or set with kcurrent)"))
        return
    indent = IndentContextManager()
    threads = []
    for task in pwndbg.commands.ktask.get_ktasks():
        threads += task.threads
    for thread in threads:
        if thread.pid != pid:
            continue
        indent.print(thread)
        with indent:
            for i, file in thread.files():
                if fd is not None and i != fd:
                    continue
                addr = int(file)
                ops = int(file["f_op"])
                prefix = indent.prefix(f"[fileno {i:03}]")
                flags = ctx_color.format_flags(int(file["f_mode"]), fmode_flags)
                desc = f"ops @ {color.red(pwndbg.chain.format(ops, limit=0))}"
                indent.print(f"- {prefix} file @ {indent.addr_hex(addr)}: {desc}")
                private_data = int(file["private_data"])
                with indent:
                    indent.print(f"private: {indent.addr_hex(private_data)}, fmode: {flags}")


parser = argparse.ArgumentParser(
    description="""
    Displays the current kernel task debugged by the debugger (gdb/lldb) if pid == None
    Displays the task with pid if pid != None.
    """
)
parser.add_argument("pid", nargs="?", type=int, help="")
parser.add_argument(
    "--set",
    dest="set_pid",
    action="store_true",
    help="sets the kernel task used for supported pwndbg commands (kfile, pagewalk, vmmap), this option does not change internal mem (purely effects how certain commands behaves)",
)


@pwndbg.commands.Command(parser, category=pwndbg.commands.CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWhenPagingEnabled
@pwndbg.commands.OnlyWithKernelDebugInfo
def kcurrent(pid=None, set_pid=False, verbose=True):
    global KCURRENT_PID, KCURRENT_PGD
    kthread = None
    if pid is None:
        kcurrent = pwndbg.aglib.kernel.current_task()
        kcurrent = pwndbg.aglib.memory.get_typed_pointer("struct task_struct", kcurrent)
        if kcurrent and pwndbg.aglib.memory.is_kernel(int(kcurrent)):
            pid = int(kcurrent["pid"])
    if pid is not None:
        for task in pwndbg.commands.ktask.get_ktasks():
            for _kthread in task.threads:
                if _kthread.pid == pid:
                    kthread = _kthread
    if kthread is None:
        print(message.warn("cannot find kernel task"))
        return
    if verbose:
        indent.print(kthread)
    if set_pid:
        mm = kthread.mm
        if not mm:
            print(message.warn("mm not found, current kernel task not set."))
            return
        KCURRENT_PID = pid
        KCURRENT_PGD = int(mm["pgd"])



from __future__ import annotations

import pwndbg.aglib
import pwndbg.aglib.proc
import pwndbg.commands
from pwndbg.commands import CommandCategory


@pwndbg.commands.Command(
    "Prints out the FS base address. See also $fsbase.", category=CommandCategory.REGISTER
)
@pwndbg.commands.OnlyWhenRunning
@pwndbg.aglib.proc.OnlyWithArch(["i386", "x86-64"])
def fsbase() -> None:
    """
    Prints out the FS base address. See also $fsbase.
    """
    print(hex(int(pwndbg.aglib.regs.fsbase)))


@pwndbg.commands.Command(
    "Prints out the GS base address. See also $gsbase.", category=CommandCategory.REGISTER
)
@pwndbg.commands.OnlyWhenRunning
@pwndbg.aglib.proc.OnlyWithArch(["i386", "x86-64"])
def gsbase() -> None:
    """
    Prints out the GS base address. See also $gsbase.
    """
    print(hex(int(pwndbg.aglib.regs.gsbase)))


# See pwndbg.gdblib.functions for the $fsbase() and $gsbase() definitions.



"""
Hexdump implementation, ~= stolen from pwntools.
"""

from __future__ import annotations

import math
import string

import pwnlib.util.lists

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.color.hexdump as H
import pwndbg.dbg_mod
from pwndbg.color import theme
from pwndbg.commands.windbg import enhex

color_scheme = None
printable = None


def groupby(width: int, array, fill=None):
    return pwnlib.util.lists.group(width, array, underfull_action="fill", fill_value=fill)


config_colorize_ascii = theme.add_param(
    "hexdump-colorize-ascii", True, "whether to colorize the hexdump command ascii section"
)
config_separator = theme.add_param(
    "hexdump-ascii-block-separator", "│", "block separator char of the hexdump command"
)
config_byte_separator = theme.add_param(
    "hexdump-byte-separator",
    " ",
    "separator of single bytes in hexdump (does NOT affect group separator)",
)


@pwndbg.config.trigger(
    H.config_normal, H.config_zero, H.config_special, H.config_printable, config_colorize_ascii
)
def load_color_scheme() -> None:
    global color_scheme, printable
    #
    # We want to colorize the hex characters and only print out
    # printable values on the right hand side.
    #
    color_scheme = {i: H.normal(f"{i:02x}") for i in range(256)}
    printable = {i: H.normal(".") for i in range(256)}

    for c in bytearray(
        (string.ascii_letters + string.digits + string.punctuation).encode("utf-8", "ignore")
    ):
        color_scheme[c] = H.printable(f"{c:02x}")
        printable[c] = (
            H.printable(f"{chr(c)}") if pwndbg.config.hexdump_colorize_ascii else f"{chr(c)}"
        )

    for c in bytearray(b"\x00"):
        color_scheme[c] = H.zero(f"{c:02x}")
        printable[c] = H.zero(".") if pwndbg.config.hexdump_colorize_ascii else "."

    for c in bytearray(b"\xff\x7f\x80"):
        color_scheme[c] = H.special(f"{c:02x}")
        printable[c] = H.special(".") if pwndbg.config.hexdump_colorize_ascii else "."

    color_scheme[-1] = "  "
    printable[-1] = " "


def hexdump(
    data: bytes,
    address: int = 0,
    width: int = 16,
    group_width: int = 4,
    flip_group_endianness: bool = False,
    skip: bool = True,
    offset: int = 0,
    size: int = 0,
    count: int = 0,
    repeat: bool = False,
    dX_call: bool = False,
):
    if not dX_call:
        if not color_scheme or not printable:
            load_color_scheme()

        # If there's nothing to print, just print the offset and address and return
        if not data:
            yield H.offset(f"+{offset:04x} ") + H.address(f"{address:#08x}  ")

            # Don't allow iterating over this generator again
            return

        data = list(bytearray(data))

        # Hexdump lines to skip_lines values and yields:
        #
        # <init: skip_lines = -1>
        # line AAAA     => skip_lines =  0   => yield "AAAA"
        # line AAAA     => skip_lines =  1   => <continue>
        # line AAAA     => skip_lines = -1   => yield "skipped ..." + "AAAA"
        # line BBBB     => skip_lines = -1   => yield "BBBB"
        skip_lines = -1

        config_separator_str = H.separator(str(config_separator))
        config_byte_separator_str = str(config_byte_separator)

        groupped = groupby(width, data, fill=-1)
        before_last_idx = len(groupped) - 2

        for i, line in enumerate(groupped):
            # Handle skipping of identical lines (see skip_lines comment above)
            if skip:
                # Count lines to be skipped by checking next/future line
                if i <= before_last_idx and line == groupped[i + 1]:
                    skip_lines += 1

                    # Since we count from -1 then 0 means we are on first line
                    # We want to yield that line, so we do not continue on that counter
                    if skip_lines != 0:
                        continue

                elif skip_lines > 0:
                    out = f"... ↓            skipped {skip_lines} identical lines ({skip_lines * width} bytes)"
                    skip_lines = -1
                    yield out
                    # Fallthrough (do not continue) so we yield the current line too

            increment = i * width
            hexline = [
                H.offset(f"+{offset + increment:04x} "),
                H.address(f"{address + increment:#08x}  "),
            ]

            for group in groupby(group_width, line):
                group = reversed(group) if flip_group_endianness else group
                for idx, char in enumerate(group):
                    if flip_group_endianness and idx == group_width - 1:
                        hexline.append(H.highlight_group_lsb(color_scheme[char]))
                    else:
                        hexline.append(color_scheme[char])
                    hexline.append(config_byte_separator_str)
                hexline.append(" ")

            hexline.append(config_separator_str)
            for group in groupby(group_width, line):
                for char in group:
                    hexline.append(printable[char])
                hexline.append(config_separator_str)

            yield "".join(hexline)

    else:
        # Traditionally, windbg will display 16 bytes of data per line.
        values = []

        if repeat:
            count = hexdump.last_count
            address = hexdump.last_address
        else:
            address = int(address) & pwndbg.aglib.arch.ptrmask
            count = int(count)

        size_type = pwndbg.aglib.typeinfo.get_type(size)

        for i in range(count):
            try:
                gval = pwndbg.aglib.memory.get_typed_pointer_value(size_type, address + i * size)
                values.append(int(gval))
            except pwndbg.dbg_mod.Error:
                break

        if not values:
            print("Could not access the provided address")
            return

        n_rows = int(math.ceil(count * size / 16.0))
        row_sz = 16 // size
        rows = [values[i * row_sz : (i + 1) * row_sz] for i in range(n_rows)]
        lines = []

        for i, row in enumerate(rows):
            if not row:
                continue
            line = [enhex(pwndbg.aglib.arch.ptrsize, address + (i * 16)), "   "]
            for value in row:
                line.append(enhex(size, value))
            lines.append(" ".join(line))

        hexdump.last_count = count
        hexdump.last_address = address + len(rows) * 16

        yield lines



from __future__ import annotations


# common functions
def hex2ptr_common(arg: str) -> int:
    """Converts a hex string to a little-endian integer address."""
    arg = "".join(filter(str.isalnum, arg))
    if len(arg) % 2 != 0:
        raise ValueError("Hex string must contain an even number of characters.")
    try:
        big_endian_num = int(arg, 16)
        num_bytes = big_endian_num.to_bytes((len(arg) + 1) // 2, byteorder="big")
        little_endian_num = int.from_bytes(num_bytes, byteorder="little")
    except ValueError as e:
        raise ValueError(f"Invalid hex string: {e}")
    return little_endian_num



from __future__ import annotations

import itertools
import struct
from collections.abc import Callable
from collections.abc import Generator
from typing import Generic
from typing import TypeVar

import pwndbg
import pwndbg.aglib.memory
import pwndbg.aglib.symbol


def _uleb128(ptr: int) -> tuple[int, int]:
    """
    Decode a ULEB128 value at the start of the given address, and return the
    decoded number, along with how many bytes the entire number takes.
    """
    acc = 0
    i = 0
    while True:
        byte = pwndbg.aglib.memory.u8(ptr + i)

        acc |= (byte & 0x7F) << (7 * i)
        if byte & 0x80 == 0:
            # This is the terminator byte.
            break

        i += 1

    return acc, i + 1


class _RawTrie:
    """
    This is the untyped base implementation of Trie.
    """

    def __init__(self, ptr: int):
        self._ptr = ptr

    def _walk(
        self,
        offset: int,
        acc: bytes,
        edgesel: Callable[[bytes, bytes], bool],
        nodesel: Callable[[bytes], bool],
    ) -> Generator[tuple[bytes, int, int]]:
        """
        Walk the trie.

        Allows callers to select edges for exploration and nodes for yielding
        through the `edgesel` and `nodesel` callables.

        At every edge, this function will call `edgesel` with the currently
        accumulated name and the name associated with the edge, and will take
        action according to the value it returns. If it returns True, that edge
        will be explored, otherwise, the edge will be ignored.

        At every node, this function will call `nodesel` with the currently
        accumulated name. If it returns True, the node will be yielded,
        otherwise, it will be ignored.

        Yielded node information consists of a tuple of (name, ptr, length),
        where `name` is the name of the node, `ptr` is the address of the first
        byte of its associated data, and `length` is the length of its
        associated data, in bytes.
        """
        base = self._ptr + offset

        node_data_len, node_data_len_len = _uleb128(base)
        if node_data_len != 0 and nodesel(acc):
            # The user selected this node, stop the walk here.
            yield acc, base + node_data_len_len, node_data_len

        cursor = base + node_data_len_len + node_data_len

        # The number of children is NOT a ULEB128.
        children = pwndbg.aglib.memory.u8(cursor)
        cursor += 1

        for _ in range(children):
            name = pwndbg.aglib.memory.string(cursor)
            cursor += len(name) + 1

            child_offset, child_offset_len = _uleb128(cursor)
            cursor += child_offset_len

            if edgesel(acc, name):
                yield from self._walk(child_offset, acc + name, edgesel, nodesel)

            # The cursor is already at the next child.

    def _get_raw(self, name: bytes) -> tuple[bytes, int, int] | None:
        """
        Get the data associated with the node of given name, if it exists.
        """

        def nodesel(candidate: bytes) -> bool:
            return candidate == name

        def edgesel(acc: bytes, candidate: bytes) -> bool:
            return name[len(acc) :].startswith(candidate)

        return next(self._walk(0, b"", edgesel, nodesel), None)

    def _entries_raw(self) -> Generator[tuple[bytes, int, int]]:
        """
        List all the entries in the trie, along with their associated data.
        """
        yield from self._walk(0, b"", lambda _acc, _candidate: True, lambda _candidate: True)

    def keys(self) -> Generator[bytes]:
        """
        List the name of all nodes in the trie.
        """
        yield from (name for name, _ptr, _size in self._entries_raw())


T = TypeVar("T")


class Trie(_RawTrie, Generic[T]):
    """
    Prefix Tree

    The Mach-O format makes extensive use of prefix trees for any operation that
    involves string-based loookup.
    """

    def __init__(self, ptr: int, ty: Callable[[int, int], T]):
        super().__init__(ptr)
        self._ty = ty

    def get(self, name: bytes) -> T | None:
        """
        Get the data associated with the node of given name, if it exists.
        """
        _, ptr, size = self._get_raw(name)
        return self._ty(ptr, size)

    def entries(self) -> Generator[tuple[bytes, T]]:
        """
        List all the entries in the trie, along with their associated data.
        """
        yield from ((name, self._ty(ptr, size)) for name, ptr, size in self._entries_raw())


def _uleb128_ty(ptr: int, size: int) -> int:
    "The type function of ULEB128 associated data, for use with Trie"

    value, actual_size = _uleb128(ptr)

    # Can fail if the type is wrong or the trie is corrupted.
    assert size == actual_size, "Size mismatch while validating ULEB128"

    return value


class DyldSharedCacheMapping:
    def __init__(self, addr: int, size: int, file_offset: int, max_prot: int, init_prot: int):
        self.addr = addr
        self.size = size
        self.file_offset = file_offset
        self.max_prot = max_prot
        self.init_prot = init_prot


def _lookup8(blob: bytes, level: int) -> int:
    """
    Hashes a variable-length byte array into a 64-bit integer.

    Apple uses a variation of an algorithm published by Bob Jenkins in 1997 on
    Dr. Dobb's Journal, and later republished on their website under the title
    "The Hash"[1]. The version used by Apple was also written by Jenkins[2], but
    does not seem to be mentioned in any of their articles, so I don't couldn't
    gather much information about it besides that it looks like a 64-bit variant
    of the algorithm in the article.

    This function is a direct Python port of the algorithm in [2].

    [1]: https://burtleburtle.net/bob/hash/doobs.html
    [2]: https://burtleburtle.net/bob/c/lookup8.c
    """
    blob = bytearray(blob)
    orig_len = len(blob)

    a = level
    b = level
    c = 0x9E3779B97F4A7C13

    padded = False
    while True:
        blob_len = len(blob)
        if blob_len == 0:
            if not padded:
                # We need to mix one more time if the blob was not padded.
                c += orig_len
                a, b, c = _mix64(a, b, c)

            break

        if blob_len < 24:
            # If the length of the blob is not divisible by 24, we pad it out
            # with zeroes until it is.
            #
            # We must be careful so as to always insert a zero at index 16,
            # which corresponds with the reservation of the length in `c` in the
            # original C code.
            c += orig_len

            blob.extend(b"\0" * (23 - blob_len))
            blob.insert(16, 0)

            padded = True

        a += (
            blob[0]
            + (blob[1] << 8)
            + (blob[2] << 16)
            + (blob[3] << 24)
            + (blob[4] << 32)
            + (blob[5] << 40)
            + (blob[6] << 48)
            + (blob[7] << 56)
        )
        b += (
            blob[8]
            + (blob[9] << 8)
            + (blob[10] << 16)
            + (blob[11] << 24)
            + (blob[12] << 32)
            + (blob[13] << 40)
            + (blob[14] << 48)
            + (blob[15] << 56)
        )
        c += (
            blob[16]
            + (blob[17] << 8)
            + (blob[18] << 16)
            + (blob[19] << 24)
            + (blob[20] << 32)
            + (blob[21] << 40)
            + (blob[22] << 48)
            + (blob[23] << 56)
        )

        a %= 0x10000000000000000
        b %= 0x10000000000000000
        c %= 0x10000000000000000

        a, b, c = _mix64(a, b, c)

        blob = blob[24:]

    return c


def _mix64(a: int, b: int, c: int) -> tuple[int, int, int]:
    """
    Mix 3 64-bit values reversibly.

    This function is part of the Python port of Bob Jenkin's hash algorithm, as
    detailed in `_lookup8`.
    """
    a -= b
    a -= c
    a ^= c >> 43
    a %= 0x10000000000000000

    b -= c
    b -= a
    b ^= a << 9
    b %= 0x10000000000000000

    c -= a
    c -= b
    c ^= b >> 8
    c %= 0x10000000000000000

    a -= b
    a -= c
    a ^= c >> 38
    a %= 0x10000000000000000

    b -= c
    b -= a
    b ^= a << 23
    b %= 0x10000000000000000

    c -= a
    c -= b
    c ^= b >> 5
    c %= 0x10000000000000000

    a -= b
    a -= c
    a ^= c >> 35
    a %= 0x10000000000000000

    b -= c
    b -= a
    b ^= a << 49
    b %= 0x10000000000000000

    c -= a
    c -= b
    c ^= b >> 11
    c %= 0x10000000000000000

    a -= b
    a -= c
    a ^= c >> 12
    a %= 0x10000000000000000

    b -= c
    b -= a
    b ^= a << 18
    b %= 0x10000000000000000

    c -= a
    c -= b
    c ^= b >> 22
    c %= 0x10000000000000000

    return a, b, c


class DyldSharedCacheHashSet:
    """
    A hash set from the DyLD Shared Cache.

    The DyLD Shared Cache uses hash sets in all structures related to Objective-C
    Optimization. This class is an interface to them.
    """

    def __init__(self, ptr: int):
        self._ptr = ptr

        self.capacity = pwndbg.aglib.memory.u32(self._ptr + 0x04)
        self.shift = pwndbg.aglib.memory.u32(self._ptr + 0x0C)
        self.mask = pwndbg.aglib.memory.u32(self._ptr + 0x10)
        self.salt = pwndbg.aglib.memory.u64(self._ptr + 0x18)

        # Mask must always be one minus a power of two. If this fails, it hints
        # that we loaded from an invalid address.
        assert (self.mask + 1).bit_count() == 1

        # Name the offsets of elements in the dynamically-sized portion of the
        # structure (which starts at 0x420).
        self._checkbytes_offset = 0x420 + self.mask + 1
        self._offsets_offset = self._checkbytes_offset + self.capacity

        # Preload the scramble and tab lists, to save on LLDB calls later on.
        self._scramble = pwndbg.aglib.memory.read(self._ptr + 0x20, 0x400)
        self._tab = pwndbg.aglib.memory.read(self._ptr + 0x420, self.mask + 1)

        # It is possible that the offsets array is not aligned. The code in
        # libmacho does not seem to care about this condition, but we should
        # probably watch out if it ever does arise in a real-world scenario.
        assert self._offsets_offset % 4 == 0, "Unaligned offset array in Mach-O perfect hash map"

    def _index_of(self, key: bytes) -> int:
        lookup = _lookup8(key, self.salt)

        tab = lookup & self.mask
        tabbed = self._tab[tab]

        scrambled = struct.unpack("<I", self._scramble[tabbed * 4 : (tabbed + 1) * 4])[0]

        return ((lookup >> self.shift) % 0x100000000) ^ scrambled

    def lookup(self, key: bytes) -> int | None:
        """
        Look up the given key in the hash set.

        Returns a pointer to the key if it is present, None otherwise.
        """
        index = self._index_of(key)

        # In libmacho, Apple uses the checkbytes as a way to quickly reject
        # elements that are not in the list without having to compare the keys,
        # but we currently have no need for that optimization.
        offset = pwndbg.aglib.memory.s32(self._ptr + self._offsets_offset + index * 4)
        if offset == 0:
            return None

        ptr = self._ptr + offset

        val = pwndbg.aglib.memory.string(ptr)
        if val != key:
            return None

        return ptr

    def keys(self) -> Generator[bytes]:
        """
        Returns an iterator over all the keys present in the hash set.
        """
        for i in range(self.capacity):
            offset = pwndbg.aglib.memory.s32(self._ptr + self._offsets_offset + i * 4)
            if offset == 0:
                continue

            yield pwndbg.aglib.memory.string(self._ptr + offset)


class DyldSharedCache:
    """
    Handle to the DyLD Shared Cache in the address space of the inferior.

    The shared cache format handling code in libmacho has multiple paths for
    gathering the same information, depending on a value that is near the
    beggining of the header, which indicates that the format has likely evolved
    quite a bit since its first intoduction.

    The way the version of a given shared cache is determined isn't exactly
    straighforward, and relies on a combination of the `magic` and
    `mappingOffset` values. Fortunately for us, however, when `mappingOffset` is
    used for this purpose, it follows the fairly widely used pattern of using
    the size of the struct to denote its version.
    """

    slide: int
    "The slide value of the DyLD Shared Cache, in bytes."

    def __init__(self, addr: int):
        self.addr = addr

        # Preload a few a few values, to speed things up later.
        self.slide = self._slide()
        images_offset = 0x18 if self._header_size() <= 0x1C4 else 0x1C0
        self._images_base = self.addr + pwndbg.aglib.memory.u32(self.addr + images_offset)
        self.image_count = pwndbg.aglib.memory.u32(self.addr + images_offset + 4)

        # Check whether the images are sorted by loading address.
        self._images_sorted_by_address = all(
            a[1] <= b[1] for a, b in itertools.pairwise(self.images)
        )

    def _header_size(self) -> int:
        """
        The length of the shared cache header, in bytes.
        """
        # Read `mappingOffset` (+0x10) from the structure.
        return pwndbg.aglib.memory.u32(self.addr + 16)

    def mappings(self) -> Generator[DyldSharedCacheMapping]:
        """
        Generate the list of memory mappings in the shared cache.
        """
        if self._header_size() <= 0x138:
            # This header predates `mappingWithSlideOffset` (+0x138), so use the
            # regular `mappingOffset` value and regular mapping structures. Read
            # the number of mapping structures from `mappingCount` (+0x14).
            base = self.addr + self._header_size()
            count = pwndbg.aglib.memory.u32(self.addr + 0x14)

            for i in range(count):
                entry = base + i * 0x20
                yield DyldSharedCacheMapping(
                    pwndbg.aglib.memory.u64(entry),
                    pwndbg.aglib.memory.u64(entry + 8),
                    pwndbg.aglib.memory.u64(entry + 16),
                    pwndbg.aglib.memory.u32(entry + 24),
                    pwndbg.aglib.memory.u32(entry + 28),
                )
        else:
            # We can use `mappingWithSlideOffset` (+0x138) and mapping with
            # slide structures for the mappings. Read the number of mapping
            # structures from `mappingWithSlideCount` (+0x13c).
            base = self.addr + pwndbg.aglib.memory.u32(self.addr + 0x138)
            count = pwndbg.aglib.memory.u32(self.addr + 0x13C)

            for i in range(count):
                entry = base + i * 0x38
                yield DyldSharedCacheMapping(
                    pwndbg.aglib.memory.u64(entry),
                    pwndbg.aglib.memory.u64(entry + 8),
                    pwndbg.aglib.memory.u64(entry + 16),
                    pwndbg.aglib.memory.u32(entry + 48),
                    pwndbg.aglib.memory.u32(entry + 52),
                )

    @property
    def base(self) -> int:
        """
        The base virtual address of the DyLD Shared Cache.
        """
        return self.addr

    @property
    def size(self) -> int:
        """
        The mapped size, in bytes, of the DyLD Shared Cache.
        """
        if self._header_size() >= 0x18C:
            # Use `sharedRegionSize` (+0xe8) as the size of the entire shared
            # region.
            return pwndbg.aglib.memory.u64(self.addr + 0xE8)
        # Find the smallest region that covers all the mappings as the size.
        start = None
        end = None
        for mapping in self.mappings():
            if start is None or start > mapping.addr:
                start = mapping.addr

            this_end = start + mapping.size
            if end is None or end < this_end:
                end = this_end

        # Technically possible, but more likely indicates that we messed up
        # somewhere along the line when interpreting mapping information.
        assert start is not None and end is not None, "No dyld shared cache mappings?"
        assert end >= start

        return end - start

    def _slide(self) -> int:
        "The slide value of the DyLD Shared Cache, in bytes."
        mapping_ptr = self.base + self._header_size()
        mapping_base = pwndbg.aglib.memory.u64(mapping_ptr)

        # Make sure this is the start of the shared cache.
        #
        # Again, technically possible, but this breaks compatibility in a way
        # that we have no idea how to deal with. Better to fail and figure out
        # we're doing something wrong than have to track a random bug back to
        # this point.
        mapping_fileoff = pwndbg.aglib.memory.u64(mapping_ptr + 0x10)
        assert mapping_fileoff == 0, (
            "First mapping of the shared cache is not at the start of the shared cache"
        )

        slide = self.base - mapping_base
        assert slide >= 0, "Slide value is negative, but we don't expect it to be"

        return slide

    @property
    def image_index_trie(self) -> Trie[int] | None:
        """
        The trie of image indices, if available.
        """
        if self._header_size() <= 0x110:
            return None

        trie_unslid = pwndbg.aglib.memory.u64(self.addr + 0x108)
        trie_ptr = trie_unslid + self.slide

        return Trie(trie_ptr, _uleb128_ty)

    def image_base(self, index: int):
        assert self.image_count > index

        return pwndbg.aglib.memory.u64(self._images_base + index * 0x20) + self.slide

    def image_name(self, index: int):
        assert self.image_count > index

        return pwndbg.aglib.memory.string(
            self.addr + pwndbg.aglib.memory.u32(self._images_base + index * 0x20 + 0x18)
        )

    @property
    def images(self) -> Generator[tuple[bytes, int]]:
        # This is a little convoluted, but this function is quite hot and
        # calling the debugger can be quite slow, so pulling in the whole array
        # at once goes a really long way.
        #
        # Yes, even with the extra logic. Python is slow, but it's not as
        # slow as calling LLDB an extra time on every iteration.
        data = pwndbg.aglib.memory.read(self._images_base, 0x20 * self.image_count)

        for i in range(self.image_count):
            base = i * 0x20
            yield (
                pwndbg.aglib.memory.string(
                    self.addr + struct.unpack("<I", data[base + 0x18 : base + 0x1C])[0]
                ),
                struct.unpack("<Q", data[base : base + 8])[0] + self.slide,
            )

    @property
    def images_sorted(self) -> Generator[tuple[bytes, int]]:
        "Same as images, but guaranteed to be sorted by increasing base address"
        if self._images_sorted_by_address:
            # The images are naturally sorted by increasing base address.
            #
            # This should be true the _vast_ majority of the time, and perhaps
            # even all the time. Just connect the generators.
            yield from self.images
        else:
            # The images are sorted in some other order.
            #
            # This should be very rare, but we shoulnd't fail if it happens.
            # Unlike the other cases in which we have to choose whether to fail
            # at or gracefully handle a weird condition, libmacho doesn't seem
            # to rely on this being the case.
            images = list(self.images)
            images.sort(key=lambda image: image[1])

            yield from iter(images)

    def is_address_in_shared_cache(self, addr: int) -> int:
        """
        Whether the given address is in the shared cache.
        """
        return addr >= self.base and addr < self.base + self.size

    def objc_builtin_selectors(self) -> DyldSharedCacheHashSet:
        """
        Looks up the hash table of builtin Objective-C selectors and returns it.
        """
        if self._header_size() > 0x1D8:
            # Use `objcOptsOffset` and the new Objective-C optimizations header
            # to find the address of the symbol hash set.

            objc_opt_offset = pwndbg.aglib.memory.u64(self.addr + 0x1D0)
            objc_opt_ptr = self.addr + objc_opt_offset

            offset = pwndbg.aglib.memory.u64(objc_opt_ptr + 0x18)
            ptr = self.addr + offset

            # Technically possible, but we have *no* idea what to do if this
            # happens, and it's more likely that we got something wrong.
            assert offset != 0, (
                "Tried to query builtin selector identity, but have no Objective-C optimization header?"
            )
        else:
            raise NotImplementedError(
                "Objective-C optimization queries are not yet supported for shared caches that have no objcOptsOffset value"
            )

        return DyldSharedCacheHashSet(ptr)


_global_new_variable_id = 0


@pwndbg.lib.cache.cache_until("objfile")
def shared_cache() -> DyldSharedCache | None:
    """
    Base address of the Darwin shared cache.

    In Darwin, the way the Objective-C Runtime queries for this value is to call
    `_dyld_get_shared_cache_range` from libdyld[1], which then calls a routine
    that lives inside dyld itself, and that returns the values after poking into
    internal C++ structures.

    From our perspective, that kind of sucks. Calling routines from debuggers
    can be quite unreliable, and so ideally we'd always be peeking into the data
    structures directly. But, in this case, even for Apple these are considered
    entirely private to dyld[2], and so there's even less of a stability guarantee
    for the layout of these structures than normal.

    Because of this, a level of care must be taken before calling this function,
    as it must be assumed that the state of the inferior can be changed by it.

    [1]: https://github.com/apple-oss-distributions/objc4/blob/f126469408dc82bd3f327217ae678fd0e6e3b37c/runtime/objc-opt.mm#L434
    [2]: https://github.com/apple-oss-distributions/dyld/blob/main/doc/dyld4.md#libdylddylib
    """
    if pwndbg.aglib.symbol.lookup_symbol("_dyld_get_shared_cache_range") is None:
        return None

    # Due to bug: https://github.com/llvm/llvm-project/issues/84806#issuecomment-1995055683
    # we have to create new variable on each call
    global _global_new_variable_id
    _global_new_variable_id += 1
    var = f"$_pwndbg_internal_shared_cache_size{_global_new_variable_id}"

    base = pwndbg.dbg.selected_inferior().evaluate_expression(
        f"size_t {var} = 0; (const void*)_dyld_get_shared_cache_range(&{var})"
    )
    if base.is_optimized_out:
        return None

    base = int(base)
    if base == 0:
        return None

    return DyldSharedCache(base)



"""
Library for non-debugger-dependant functionality.
"""

from __future__ import annotations

from pwndbg.lib.err import ErrorCode
from pwndbg.lib.err import Status
from pwndbg.lib.err import SymbolNotRecoveredError
from pwndbg.lib.err import TypeNotFoundError
from pwndbg.lib.err import TypeNotRecoveredError

__all__ = [
    "ErrorCode",
    "Status",
    "TypeNotRecoveredError",
    "TypeNotFoundError",
    "SymbolNotRecoveredError",
]



"""
Find a chain of leaks given some starting address.
"""

from __future__ import annotations

import argparse
import queue

import pwndbg
import pwndbg.aglib.memory
import pwndbg.aglib.qemu
import pwndbg.aglib.vmmap
import pwndbg.chain
import pwndbg.color.memory as mem_color
import pwndbg.commands
import pwndbg.dbg_mod
from pwndbg.color import message
from pwndbg.commands import CommandCategory


# Used to recursively print the pointer chain.
# addr is a pointer. It is taken to be a child pointer.
# visited_map is a map of children -> (parent,parent_start)
def get_rec_addr_string(addr, visited_map):
    page = pwndbg.aglib.vmmap.find(addr)
    arrow_right = pwndbg.chain.c.arrow(f" {pwndbg.config.chain_arrow_right} ")

    if page is not None:
        if addr not in visited_map:
            return ""

        parent_info = visited_map[addr]
        parent = parent_info[0]
        parent_base_addr = parent_info[1]
        if parent - parent_base_addr < 0:
            curText = hex(parent_base_addr) + hex(parent - parent_base_addr)
        else:
            curText = hex(parent_base_addr) + "+" + hex(parent - parent_base_addr)
        if parent_base_addr == addr:
            return ""
        return (
            get_rec_addr_string(parent_base_addr, visited_map)
            + mem_color.get(parent_base_addr, text=curText)
            + arrow_right
        )
    return ""


# Useful for debugging. Prints a map of child -> (parent, parent_start)
def dbg_print_map(maps) -> None:
    for child, parent_info in maps.items():
        print(f"0x{child:x} + (0x{parent_info[0]:x}, 0x{parent_info[1]:x})")


parser = argparse.ArgumentParser(
    description="""
Attempt to find a leak chain given a starting address.

Scans memory near the given address, looks for pointers, and continues that process to attempt to find leaks.
""",
)
parser.add_argument(
    "address",
    type=pwndbg.commands.AddressExpr,
    nargs="?",
    default="$sp",
    help="Starting address to find a leak chain from",
)
parser.add_argument(
    "-p",
    "--page_name",
    type=str,
    nargs="?",
    default=None,
    help="Substring required to be part of the name of any found pages",
)
parser.add_argument(
    "-o",
    "--max_offset",
    type=int,
    default=0x48,
    nargs="?",
    help="Max offset to add to addresses when looking for leak",
)
parser.add_argument(
    "-d",
    "--max_depth",
    type=int,
    default=0x4,
    nargs="?",
    help="Maximum depth to follow pointers to",
)
parser.add_argument(
    "-s",
    "--step",
    nargs="?",
    type=int,
    default=0x1,
    help="Step to add between pointers so they are considered. For example, if this is 4 it would only consider pointers at an offset divisible by 4 from the starting pointer",
)
parser.add_argument(
    "--negative_offset",
    nargs="?",
    type=int,
    default=0x0,
    help="Max negative offset to search before an address when looking for a leak",
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MEMORY,
    examples="""
pwndbg> leakfind $rsp --page_name=filename --max_offset=0x48 --max_depth=6.
This would look for any chains of leaks that point to a section in filename
which begin near $rsp, are never 0x48 bytes further from a known pointer,
and are a maximum length of 6.
""",
)
@pwndbg.commands.OnlyWhenRunning
def leakfind(
    address=None,
    page_name=None,
    max_offset: int = 0x40,
    max_depth: int = 0x4,
    step: int = 0x1,
    negative_offset: int = 0x0,
):
    if address is None:
        print("No start address provided.")
        return

    if not pwndbg.aglib.memory.peek(address):
        print("Unable to read from starting address.")
        return

    # Just warn the user that a large depth might be slow.
    # Probably worth checking offset^depth < threshold. Do this when more benchmarking is established.
    if max_depth > 8:
        print(message.warn("leakfind may take a while to run on larger depths."))

    stride = step

    # The below map stores a map of child address->(parent_address,parent_start_address)
    # In the above tuple, parent_address is the exact address with a pointer to the child address.
    # parent_start_address is an address that a previous address pointed to.
    # We need to store both so that we can nicely create our leak chain.
    visited_map = {}
    visited_set = {address}
    address_queue: queue.Queue[int] = queue.Queue()
    address_queue.put(address)
    depth = 0
    time_to_depth_increase = 0

    # Run a bfs
    # TODO look into performance gain from checking if an address is mapped before calling pwndbg.aglib.memory.pvoid()
    # TODO also check using pwndbg.aglib.memory.read for possible performance boosts.
    while address_queue.qsize() > 0 and depth < max_depth:
        if time_to_depth_increase == 0:
            depth = depth + 1
            time_to_depth_increase = address_queue.qsize()
        cur_start_addr = address_queue.get()
        time_to_depth_increase -= 1
        for cur_addr in range(
            cur_start_addr - negative_offset, cur_start_addr + max_offset, stride
        ):
            try:
                cur_addr &= pwndbg.aglib.arch.ptrmask
                result = int(pwndbg.aglib.memory.read_pointer_width(cur_addr))
                if result in visited_map or result in visited_set:
                    continue
                visited_map[result] = (
                    cur_addr,
                    cur_start_addr,
                )  # map is of form child->(parent,parent_start)
                address_queue.put(result)
                visited_set.add(result)
            except pwndbg.dbg_mod.Error:
                # That means the memory was unmapped. Just skip it if we can't read it.
                break

    # A map of length->list of lines. Used to let us print in a somewhat nice manner.
    output_map: dict[int, list[str]] = {}
    arrow_right = pwndbg.chain.c.arrow(f" {pwndbg.config.chain_arrow_right} ")

    for child in visited_map:
        child_page = pwndbg.aglib.vmmap.find(child)
        if child_page is not None:
            if page_name is not None and page_name not in child_page.objfile:
                continue
            line = (
                get_rec_addr_string(child, visited_map)
                + mem_color.get(child)
                + " "
                + mem_color.get(child, text=child_page.objfile)
            )
            chain_length = line.count(arrow_right)
            if chain_length in output_map:
                output_map[chain_length].append(line)
            else:
                output_map[chain_length] = [line]

    # Output sorted by length of chain
    for chain_length, lines in output_map.items():
        for line in lines:
            print(line)

    if pwndbg.aglib.qemu.is_qemu():
        print("\n[QEMU target detected - leakfind result might not be accurate; see `help vmmap`]")
